/*
 * ar3cmonitor - analoggraph.cpp
 *
 * Copyright (C) 2004  Szymon Acedanski <accek@podlesie.net>
 * Copyright (C) 2005..2011  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <qpainter.h>
#include <cmath>

#include "ar3cmonitor.h"
#include "analoggraph.h"
#include "consts.h"
#include "commobj.h"
#include "control.h"
#include <time.h>
#include "analogcache.h"
#include <niceformat.h>
#include <profile.h>
#include <tscale.h>
#include <compiler.h>

#ifndef FORCE_FLOAT_ARITH
#define INTEGER_ARITH
#endif

AnalogGraph::AnalogGraph(QWidget * parent, const char *name, Qt::WFlags f)
		: BufferedWidget(parent, name, f)
{
	FirstPlot = LastPlot = NULL;
	hgridLines = NULL;

	setMinimumHeight(10);
	setMinimumWidth(100);

	cursor2 = cursor = -1;
}

AnalogGraph::~AnalogGraph()
{
	delete []hgridLines;

	AnalogPlot *p = FirstPlot;
	while (p) {
		AnalogPlot *n = p->Next;
		delete p->c;
		delete p;
		p = n;
	}
}

void AnalogGraph::mouseDoubleClickEvent(QMouseEvent *)
{
	mpressed = 0;
	cursor2 = cursor = -1;
	update();
}

void AnalogGraph::mousePressEvent(QMouseEvent *e)
{
	int pri_cursor, sec_cursor;

	switch (CO->cursor_mode_get()) {
	case CURSOR_MODE_RLM:
		pri_cursor = Qt::RightButton;
		sec_cursor = Qt::LeftButton;
		break;
	case CURSOR_MODE_MLR:
		pri_cursor = Qt::MidButton;
		sec_cursor = Qt::RightButton;
		break;
	case CURSOR_MODE_LRM:
	default:
		pri_cursor = Qt::LeftButton;
		sec_cursor = Qt::RightButton;
		break;
	}
	if (e->button() & pri_cursor) {
		mpressed = 1;
		cursor = e->x();
		update();
	} else if (e->button() & sec_cursor) {
		mpressed = 2;
		cursor2 = e->x();
		if (cursor == -1)
			cursor = cursor2;
		update();
	} else if (CO->reviewMode()) {
		mpressed = 3;
		old_pos = e->x();
	}
}

void AnalogGraph::mouseMoveEvent(QMouseEvent *e)
{
	switch (mpressed) {
		case 1:
			cursor = e->x();
			update();
			break;
		case 2:
			cursor2 = e->x();
			update();
			break;
		case 3: {
			int pos = e->x();
			int w = width();

			BUG_ON(!w);

			CO->advanceCursor((old_pos - pos) * (long long)
					CO->timeRange()	/ w);
			old_pos = pos;
			break;
		}
	}
}

void AnalogGraph::mouseReleaseEvent(QMouseEvent *)
{
	mpressed = 0;
}

void AnalogGraph::wheelEvent(QWheelEvent *e)
{
	int q = e->delta() / (8 * 15);

	if (q > 0) {
		if (CO->timeRange() < 3 * 60 * 60 * 1000000000ULL)

		if (CO->reviewMode())
			CO->advanceCursor(CO->timeRange() / 4);
		CO->setTimeRange(CO->timeRange() * 3 / 2);
	} else if (CO->timeRange() > 1000000000ULL) {
		CO->setTimeRange(CO->timeRange() * 2 / 3);
		if (CO->reviewMode())
			CO->advanceCursor(-(long long) CO->timeRange() / 4);
	}
}

QString AnalogGraph::niceFormat(AnalogPlot *ap, float v)
{
	char buf[32];

	niceformat(buf, sizeof(buf), ap->UnitValue, ap->Min, ap->Max, v,
			ap->prec);
	return QString(buf);
}

void AnalogGraph::draw_cursor(QPainter *p, int pos, QColor color)
{
	if (pos != -1) {
		QPen pen = p->pen();
		pen.setColor(color);
		p->setPen(pen);

		p->drawLine(pos, 0, pos, height() - 1);
	}
}

struct cursor_state {
	unsigned long long timestamp;
	int position;
	int value;
	int ok;
	int touched;
};

static void cs_init(struct cursor_state *cs, int position,
		unsigned long long T0, int w)
{
	cs->ok = 0;
	cs->touched = 0;
	cs->timestamp = 0;
	cs->value = 0;
	cs->position = position;
	if (cs->position != -1)
		cs->timestamp = calculateT(position, T0, w, CO->timeRange());
}

static void cs_setvalue(struct cursor_state *cs, int value)
{
	cs->value = value;
	cs->touched = 1;
	cs->ok = 1;
}

static void cs_reset(struct cursor_state *cs)
{
	cs->ok = 0;
	cs->touched = 0;
}

inline void AnalogGraph::calculate_cursor_value(struct cursor_state *cs, int x,
		unsigned long long oldt, int oldv, ac_elem *e)
{
	unsigned long long t;
	long long dt, dct;
	int cv;

	if ((cs->position == -1) || (cs->touched) || (x > cs->position))
		return;

	if (!e->defined) {
		cs->touched = 1;
		return;
	}

	t = e->ts;

	if (x == cs->position) {
		/* troche oszukujemy, ale nich uzytkownik ma to co widzi */
		cs_setvalue(cs, e->value);
	} else if (t == oldt) {
		cs->touched = 1;
	} else {
		dt = oldt - t;
		dct = cs->timestamp - t;

		cv = e->value;
		/*
		 * dct jest przewaznie male, a wiec przepelnienie nie powinno
		 * wystepowac
		 */
		cs_setvalue(cs, cv + (oldv - cv) * dct / dt);
	}
}

#ifdef HAVE_SYS_TIME_H

#include <sys/time.h>

static unsigned long long time_us(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return tv.tv_sec * 1000000ULL + tv.tv_usec;
}

#else

static unsigned long long time_us(void)
{
	return clock() * 1000000ULL / CLOCKS_PER_SEC;
}

#endif

/**
 * paint_hgrid - paint horizontal grid
 * @p: painter
 * @w: width
 * @h: height
 */
void AnalogGraph::paint_hgrid(QPainter *p, int w, int h)
{
	unsigned int i;

	QPen pen = p->pen();
	pen.setColor(hgridColor);
	p->setPen(pen);

	for (i = 0; i < hgridnLines; i++) {
		int y = (h * 10000 - (10000 - hgridLines[i])
				* (h - 4)) / 10000 - 2;
		p->drawLine(0, y, w, y);
	}
}

/**
 * paint_vgrid - paint vertical grid
 * @p: painter
 * @w: width
 * @h: height
 */
void AnalogGraph::paint_vgrid(QPainter *p, int w, int h,
		unsigned long long T0, unsigned long long T1)
{
	unsigned long long T;

	QPen pen = p->pen();
	pen.setStyle(Qt::DotLine);
	pen.setColor(CO->vgridColor());
	p->setPen(pen);

	unsigned long long step = CO->vgridStep();
	for (T = T0 + step - T0 % step; T <= T1; T += step) {
		int x = tscale(T, T0, w, CO->timeRange());

		p->drawLine(x, 2, x, h - 2);
	}
}

void AnalogGraph::draw_points(QPainter *p, QPen point_pen, QPointArray pts,
		unsigned int j)
{
	unsigned int i;

	if (show_points_count && CO->offlineMode()) {
		p->setPen(point_pen);
		for (i = 0; i < j; i++) {
			int ptx = pts[i].x();
			int pty = pts[i].y();

			p->drawEllipse(ptx - 1, pty - 1, 3, 3);
		}
	}
}

void AnalogGraph::paint(QPainter *p)
{
	unsigned long long t = time_us();
	int h = height();
	int w = width();
	struct cursor_state cs1, cs2;
	int plots;

	BUG_ON(!h);
	BUG_ON(!w);

	if (!CO->haveData()) {
		p->setPen(Qt::red);
		p->drawText(0, 0, w, h, Qt::AlignCenter, tr("No data"));
		return;
	}

	if (hgridLines)
		paint_hgrid(p, w, h);

	unsigned long long T1 = CO->cursorStamp();
	unsigned long long timerange = CO->timeRange();
	unsigned long long T0 = T1 - timerange;

	draw_cursor(p, cursor, ARMON_COLOR(red));
	cs_init(&cs1, cursor, T0, w);
	draw_cursor(p, cursor2, ARMON_COLOR(green));
	cs_init(&cs2, cursor2, T0, w);

	if (CO->vgridStep())
		paint_vgrid(p, w, h, T0, T1);

	p->setFont(CO->valueFont());
	int CaptionY = 0;
	bool firstPlot = true;
	int alignment = CO->reviewMode() ? Qt::AlignRight : Qt::AlignLeft;

	unsigned long long minTimeDelta = timerange / w / TIME_MSAMPLING;
	int dh = QFontMetrics(CO->valueFont()).height();
	QPen point_pen(Qt::red, 3);

	plots = 0;
	for (AnalogPlot *ap = FirstPlot; ap; ap = ap->Next)
		plots++;

	for (AnalogPlot *ap = FirstPlot; ap; ap = ap->Next) {
		AnalogCache *ac = ap->c;

		ac->set_quantum(minTimeDelta);
		ac->set_vquantum(65536 / h / VALUE_MSAMPLING);
		ac->update();

		QPointArray pts(1000);
		int j = 0;
		float Min = ap->Min;
		float Max = ap->Max;
#ifdef INTEGER_ARITH
		long imin = (long) round(Min * ap->UnitValue);
		long imax = (long) round(Max * ap->UnitValue);
#endif
		int x0 = w;
		int counter = 0;
		struct ac_elem *e = ac->get_last();

		long A = - (h - 4);
		long B = imax - imin;
		long C = h - 2 + imin * (h - 4) / (imax - imin);

		p->setPen(ap->Color);
		if (firstPlot && (e->ts < CO->cursorStamp())) {
			int x = e->is_head ? 0 : calculateX_bin(e->ts, T0,
					w, timerange);
			p->fillRect(x + 1, 2, x0 - x - 2, h - 2,
					QBrush(CO->hgridColor(),
						Qt::DiagCrossPattern));
		}

		unsigned long long oldt = 0;
		int oldv = 0;

		cs_reset(&cs1);
		cs_reset(&cs2);

		if (!e->is_head)
			oldt = e->ts;
		while (likely(!e->is_head)) {
			unsigned long long t = e->ts;
			long long x = tscale(t, T0, w, timerange);
			counter++;

			calculate_cursor_value(&cs1, x, oldt, oldv, e);
			calculate_cursor_value(&cs2, x, oldt, oldv, e);

			oldt = t;

			if (unlikely(!e->defined)) {
				p->drawPolyline(pts, 0, j);
				draw_points(p, point_pen, pts, j);

				j = 0;
				if (t == 0) {
					t = T0;
					x = 0;
				} else if (x < 0)
					x = 0;
				else if (x > w)
					x = w;

				if (firstPlot)
					p->fillRect(x, 2, x0 - x, h - 2,
						QBrush(CO->hgridColor(),
							Qt::DiagCrossPattern));
			} else {
				int iv = e->value;
#ifndef INTEGER_ARITH
				float v = iv / ap->UnitValue;
#endif
				int tx = x;

				if (unlikely(tx < 0)) {
#ifdef INTEGER_ARITH
					/* We still using float arith
					 * for interpolation */
					float v = iv / ap->UnitValue;
#endif
					float ov = oldv / ap->UnitValue;

					v = v + (v - ov) * tx / (x0 - x);
#ifdef INTEGER_ARITH
					iv = (int) round(v * ap->UnitValue);
#endif
					tx = 0;
				} else if (unlikely(tx > w)) {
					ac_elem *e2 = e;

					e2 = list_entry(list_prev(&e2->list),
							ac_elem, list);
					if (e2->is_head) {
						e = e2;
						break;
					}

					if (!e->defined) {
						e = e2;
						continue;
					}

#ifdef INTEGER_ARITH
					/* We still using float arith
					 * for interpolation */
					float v = iv / ap->UnitValue;
#endif
					float ov = e2->value / ap->UnitValue;
					unsigned long long t2 = e2->ts;
					long long x2 = tscale(t2, T0, w,
							timerange);
					v = ov + (v - ov) * (w - x2) / (x - x2);
#ifdef INTEGER_ARITH
					iv = (int) round(v * ap->UnitValue);
#endif
					tx = w;
				}

				oldv = iv;
				if (unlikely(j == 1000)) {
					p->drawPolyline(pts, 0, j);
					draw_points(p, point_pen, pts, j);
					pts.setPoint(0, pts.point(j - 1));
					j = 1;
				}

#ifdef INTEGER_ARITH
				pts.setPoint(j++, tx, A * iv / B + C);
#else
				pts.setPoint(j++, tx, (int) (h - (v - Min)
							* (h - 4) / (Max - Min)
							- 2));
#endif
			}
			x0 = x;
			if (unlikely(t < T0)) {
				ac->cleanup(e);
				break;
			}

			e = list_entry(list_prev(&e->list), ac_elem, list);
			prefetch(e->list.prev);
		}

		QString str;
		QString str2 = "";

		if (!cs1.ok) {
			str = ap->Caption;
		} else if (cs2.ok) {
			float v = cs1.value / ap->UnitValue;
			float d = (cs1.value - cs2.value) / ap->UnitValue;
			long long dt = (cs1.timestamp - cs2.timestamp);
			float d_dt = 0;

			if (dt)
				d_dt = d * 1e9f / llabs(dt);
			if (show_points_count) {
				str = QString("%1 = %2%3 (d = %4) (%5)")
					.arg(ap->Caption)
					.arg(niceFormat(ap, v)).arg(ap->Unit)
					.arg(niceFormat(ap, d)).arg(counter);
			} else {
				str = QString("%1 = %2%3 (d = %4)")
					.arg(ap->Caption)
					.arg(niceFormat(ap, v)).arg(ap->Unit)
					.arg(niceFormat(ap, d));
			}
			str2 = QString("d%1/dt = %2")
				.arg(ap->Caption)
				.arg(niceFormat(ap, d_dt));
		} else {
			float v = cs1.value / ap->UnitValue;

			if (show_points_count) {
				str = QString("%1 = %2%3 (%6)").arg(ap->Caption)
					.arg(niceFormat(ap, v)).arg(ap->Unit)
					.arg(counter);
			} else {
				str = QString("%1 = %2%3").arg(ap->Caption)
					.arg(niceFormat(ap, v)).arg(ap->Unit);
			}

		}

		p->drawText(0, CaptionY, width() - 2, height(),
				alignment | Qt::AlignTop, str);

		if (height() > (plots + 2) * dh + CaptionY)
			p->drawText(0, 0, width() - 2, height() - CaptionY,
					alignment | Qt::AlignBottom, str2);

		CaptionY += dh;

		if (firstPlot && x0 > 0)
			p->fillRect(0, 2, x0 - 1, h - 2,
					QBrush(CO->hgridColor(),
						Qt::DiagCrossPattern));

		p->drawPolyline(pts, 0, j);
		draw_points(p, point_pen, pts, j);

		firstPlot = false;

		ap->c->untouch();
	}

	if (cs1.ok) {
		QString str;
		unsigned long long ct = cs1.timestamp;
		time_t tim = TS_SEC(ct);
		struct tm tm;
		localtime_r(&tim, &tm);

		QPen pen = p->pen();
		pen.setColor(ARMON_COLOR(white));
		p->setPen(pen);

		if (cs2.ok) {
			long long dt = ct - cs2.timestamp;

			str = QString().sprintf("t = %02d:%02d:%02d.%03d "
					"(d = %s%d.%03ds)",
					(int) tm.tm_hour,
					(int) tm.tm_min,
					(int) tm.tm_sec,
					(int) (TS_MS(ct) % 1000),
					(TS_MS(dt) >= 0) ? "" : "-",
					(int) abs(TS_SEC(dt)),
					(int) abs(TS_MS(dt) % 1000));
		} else {
			str = QString().sprintf("t = %02d:%02d:%02d.%03d",
					(int) tm.tm_hour,
					(int) tm.tm_min,
					(int) tm.tm_sec,
					(int) (TS_MS(ct) % 1000));
		}

		p->drawText(0, CaptionY, width() - 2, height(),
				alignment | Qt::AlignTop, str);
		CaptionY += dh;
	}

	if (profile) {
		qWarning("analogplot repaint time: %.3fms.",
				(time_us() - t) / 1000.0);
	}
}

void AnalogGraph::addPlot(int physical, QString caption, QColor color,
	float min, float max, float uv, QString unit, prec_t prec)
{
	if (uv == 0) {
		qWarning(tr("analoggraph: zero unit value in %1 plot")
				.arg(caption));
		return;
	}
	if (min == max) {
		qWarning(tr("analoggraph: min and max are equal in %1 plot")
				.arg(caption));
		return;
	}

	AnalogPlot *p = new AnalogPlot();
	p->Next = NULL;
	p->Physical = physical;
	p->Caption = caption;
	p->Color = color;
	p->Min = min;
	p->Max = max;
	p->UnitValue = uv;
	p->Unit = unit;
	p->prec = prec;
	p->c = new AnalogCache(physical);

	if (!FirstPlot)
		FirstPlot = LastPlot = p;
	else {
		LastPlot->Next = p;
		LastPlot = p;
	}
}

void AnalogGraph::addHGrid(int nlines, int *lines, QColor color)
{
	hgridLines = new int[nlines];
	memcpy(hgridLines, lines, nlines * sizeof(int));
	hgridnLines = nlines;
	hgridColor = color;
}

void AnalogGraph::addSizeHint(int hint)
{
	setMinimumHeight(hint);
	setMaximumHeight(hint);
}

void AnalogGraph::update_timer()
{
	if (!CO->haveData()) {
		update();
		return;
	}

	for (AnalogPlot *ap = FirstPlot; ap; ap = ap->Next) {
		AnalogCache *ac = ap->c;
		if (ac->is_touched()) {
			update();
			return;
		}
	}
}
