//
// C++ Interface: ar3cmonitor
//
// Description: 
//
//
// Author: Szymon Acedanski <accek@podlesie.net>, (C) 2004
// Copyright (C) 2005..2008  Krzysztof Mazur <krzysiek@podlesie.net>
// Copyright (C) 2008  Krzysztof Kasperski <casper@podlesie.net>
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>

#include <qlabel.h>
#include <qlayout.h>
#include <qmessagebox.h>
#include <qpushbutton.h>
#include <qmotifstyle.h>
#include <qapplication.h>
#include <qmenubar.h>
#include <qtextcodec.h>
#if QT_VERSION >= 0x40000
# include <QKeyEvent>
# include <QObject>
# include <Q3VBoxLayout>
# include <Q3HBoxLayout>
# include <QScrollArea>
# include <QInputDialog>

# define QVBoxLayout	Q3VBoxLayout
# define QHBoxLayout	Q3HBoxLayout
#else
# include <qscrollview.h>
# include <qinputdialog.h>
#endif

#include <sys/types.h>
#include <unistd.h>

#include "ar3cmonitor.h"
#include "commobj.h"
#include "meterwidget.h"
#include "consts.h"
#include "analoggraph.h"
#include "binarygraph.h"
#include "dzwonekgraph.h"
#include "ledwidget.h"
#include "timescale.h"
#include "myseparator.h"
#include "action.h"
#include "control.h"
#include "statusbar.h"
#include "wskaz.h"
#include "mytabwidget.h"
#include "ameter.h"
#include "dmeter.h"
#include "object.h"
#include "analogcache.h"
#include "about.h"
#include <hide.h>
#include <fonts.h>

#ifndef NO_SETTIME
#include "settime.h"
#endif
#ifndef NO_PENDRIVE
#include "pendrive.h"
#endif

#include "qpixmap.h"
#include "qimage.h"
#include "tclocklabel.h"

#ifdef JUMP_NODTE
# include <jumpdlg_nodte.h>
#else
# include <jumpdlg.h>
#endif

Ar3cMonitor *MainWindow;

extern int scheme;
extern int show_menu;
extern int show_statusbar;
extern int show_changesvr;
extern const char *config_encoding;
extern char *conf_locale;
extern char *password;
extern int swap_cursors_style;
extern int enable_layers;
extern int font_magnify;
extern QWidget *SS;
extern const char *ar3cmon_host;
extern int touchscreen;
extern const char *ts_calibrate;

extern const char *key_q_action;
extern const char *key_x_action;

extern int led_no_symbol;

Ar3cMonitor::Ar3cMonitor(QWidget * parent, const char *name, Qt::WFlags)
	: QWidget(parent, name)
{
	const char *family;
	unsigned int size;
	int ret;

#ifdef __SIMPAD__
	MainWindow = this;
#endif

	/*
	 * We want to have configurable fonts, but we also want to
	 * use default set by user. In most cases the default font
	 * is the right choice for normal texts. We save default
	 * font as FONT_BASE.
	 */
	QFont f;
	ar3cmon_font_set_family(AR3CMONITOR_FONT_BASE, f.family());
	ret = f.pointSize();
	if (ret == -1) {
		ret = f.pixelSize();
		if (ret == -1)
			ret = 12;
		else
			ar3cmon_font_set_pixels(1);
	}
	ret = (ret * font_magnify + 500) / 1000;
	ar3cmon_font_set_size(AR3CMONITOR_FONT_BASE, ret);

	fonts_reset();

	family = ar3cmon_font_family(AR3CMONITOR_FONT);
	if (family != NULL) {
		f.setStyleHint(QFont::SansSerif, QFont::PreferBitmap);
		f.setFamily(family);
	}

	size = ar3cmon_font_size(AR3CMONITOR_FONT);
	f.setPointSize(size);
	qApp->setFont(f, TRUE);

#if __TINYSCREEN__
	setMinimumWidth(320);
	setMinimumHeight(240);
#else
	setMinimumWidth(640);
	setMinimumHeight(480);
#endif

	CO = new ControlObject(AR3CMON_MAX_BINARY_INPUTS);
	ACC = new AnalogCacheControl(AR3CMON_MAX_ANALOG_INPUTS);

	setCaption(tr("AR-3c Monitor"));

	RootLayout = new QVBoxLayout(this);

	StatusLabel = NULL;
	RootWidget = NULL;
	ClockLabel = NULL;
	TitleLabel = NULL;
	HLayout = NULL;
	VLayout = NULL;
	VLWidget = NULL;
	CurrentStack = NULL;
	CurrentHiddenTab = NULL;
	CurrentDzwonek = NULL;
	AttrWidget = NULL;
	CurrentAnalogGraph = NULL;
	options = NULL;
	TabsCnt = 0;
	for (int i = 0; i < AR3CMON_MAX_TABS; i++)
		TabsTbl[i] = NULL;

	tabIndex = 0;

	TheCommObject = new CommObject(this);
	connect(TheCommObject, SIGNAL(sigOfflineConnected()),
			SLOT(commOfflineConnected()));
	connect(TheCommObject, SIGNAL(sigConnected()), SLOT(commConnected()));
	connect(TheCommObject, SIGNAL(sigConnecting()), SLOT(commConnecting()));
	connect(TheCommObject, SIGNAL(sigError()), SLOT(commError()));

	if (show_menu)
		configure_menu();

	BodyWidget = new QWidget(this);
	Q_CHECK_PTR(BodyWidget);
	RootLayout->addWidget(BodyWidget);

	BodyLayout = new QVBoxLayout(BodyWidget);

	createStatusLabel(tr("Initialization ..."));
	TheCommObject->connectToAr3c();
	connect(CO, SIGNAL(sigNewTimestamp()), SLOT(newTimestamp()));
	connect(CO, SIGNAL(sigPleaseUpdate()), SLOT(newTimestamp()));

	if (show_statusbar) {
		StatusBar = new MainStatusBar(this);
		RootLayout->addWidget(StatusBar);
	}

	connect(TheCommObject, SIGNAL(sigReallyConnected()),
		CO, SLOT(reallyConnected()));
}

#ifdef BUILDIN_IMAGES

#include "arrow_green_trans.xpm"
#include "arrow_red_trans.xpm"
#include "hand.xpm"

#define load_image(name) 	(new QImage(name ## _xpm))

#else
const QImage *__load_image(const char *name)
{
	return new QImage(QString(DATA_DIR) + name + ".xpm");
}

#define M_STRING(name) #name

#define load_image(name) __load_image(M_STRING(name))
#endif

void Ar3cMonitor::configure_menu(void)
{
	QWidget *t = new QWidget(this);

	QMenuBar *m = new QMenuBar(t);

	QPopupMenu *file = new QPopupMenu(t);
	file->insertItem(tr("&Connect"), TheCommObject, SLOT(connectToAr3c()));
	file->insertItem(trUtf8("&Disconnect"), TheCommObject, SLOT(goOffline()));
	if (show_changesvr) {
		file->insertSeparator();
		file->insertItem(tr("Change &Server"),
				this, SLOT(set_hostname()));
	}
	file->insertSeparator();
	file->insertItem(trUtf8("&Quit"), this, SLOT(menu_quit()));

#ifdef TIME_IN_STATUS_BAR
	file->insertSeparator();
	QPopupMenu *mode = new QPopupMenu(t);
	mode->insertItem(trUtf8("&Online"), CO, SLOT(enterOnlineMode()));
	mode->insertItem(trUtf8("&Review"), CO, SLOT(enterReviewMode()));
	mode->insertItem(trUtf8("&Offline"), CO, SLOT(enterOfflineMode()));
	file->insertItem(trUtf8("&Mode"), mode);
	file->insertSeparator();
	file->insertItem(trUtf8("&Jump"), this, SLOT(jump_clicked()));
	m->insertItem(trUtf8("&Menu"), file);
#else
	m->insertItem(trUtf8("&File"), file);
	QPopupMenu *mode = new QPopupMenu(t);
	mode->insertItem(trUtf8("&Online"), CO, SLOT(enterOnlineMode()));
	mode->insertItem(trUtf8("&Review"), CO, SLOT(enterReviewMode()));
	mode->insertItem(trUtf8("&Offline"), CO, SLOT(enterOfflineMode()));
	m->insertItem(trUtf8("&Mode"), mode);
#endif

	QPopupMenu *tr_menu = new QPopupMenu(t);
	tr_menu->insertItem(trUtf8("&Increase"), this,
			SLOT(increase_timerange()));
	tr_menu->insertItem(trUtf8("&Decrease"), this,
			SLOT(decrease_timerange()));
	tr_menu->insertSeparator();
	
	tr_menu->insertItem(trUtf8("15s"), this, SLOT(timerange_15s()));
	tr_menu->insertItem(trUtf8("30s"), this, SLOT(timerange_30s()));
	tr_menu->insertItem(trUtf8("1m"), this, SLOT(timerange_1m()));
	tr_menu->insertItem(trUtf8("2m"), this, SLOT(timerange_2m()));
	tr_menu->insertItem(trUtf8("5m"), this, SLOT(timerange_5m()));
	tr_menu->insertItem(trUtf8("10m"), this, SLOT(timerange_10m()));
	tr_menu->insertItem(trUtf8("15m"), this, SLOT(timerange_15m()));
	tr_menu->insertItem(trUtf8("30m"), this, SLOT(timerange_30m()));
	tr_menu->insertItem(trUtf8("1h"), this, SLOT(timerange_1h()));
	tr_menu->insertItem(trUtf8("2h"), this, SLOT(timerange_2h()));
	tr_menu->insertItem(trUtf8("3h"), this, SLOT(timerange_3h()));
	
	if (enable_layers) {
		tr_menu->insertSeparator();
		QPopupMenu *layer_menu = new QPopupMenu(t);
		layer_menu->insertItem(trUtf8("&Active"), this,
				SLOT(select_active_layer()));
		layer_menu->insertSeparator();
		layer_menu->insertItem(trUtf8("0"), this,
				SLOT(select_0_layer()));
		layer_menu->insertItem(trUtf8("1"), this,
				SLOT(select_1_layer()));
		layer_menu->insertItem(trUtf8("2"), this,
				SLOT(select_2_layer()));
		layer_menu->insertItem(trUtf8("3"), this,
				SLOT(select_3_layer()));
		layer_menu->insertItem(trUtf8("4"), this,
				SLOT(select_4_layer()));
		layer_menu->insertItem(trUtf8("5"), this,
				SLOT(select_5_layer()));
		layer_menu->insertItem(trUtf8("6"), this,
				SLOT(select_6_layer()));
		layer_menu->insertItem(trUtf8("7"), this,
				SLOT(select_7_layer()));
		tr_menu->insertItem(trUtf8("&Layer"), layer_menu);
	}
	
	m->insertItem(trUtf8("&Timerange"), tr_menu);

	if (swap_cursors_style == 1) {
		options = new QPopupMenu(t);
		options->setCheckable(TRUE);
		swapcur0_id = options->insertItem(trUtf8("&Primary"), this,
				SLOT(swap_cursor0()));
		swapcur1_id = options->insertItem(trUtf8("&Secondary"), this,
				SLOT(swap_cursor1()));
		swapcur2_id = options->insertItem(trUtf8("&Move"), this,
				SLOT(swap_cursor2()));
		options->setItemChecked(swapcur0_id, true);
		m->insertItem(trUtf8("&Options"), options);
	}

#ifndef TIME_IN_STATUS_BAR
	m->insertItem(trUtf8("&Jump"), this, SLOT(jump_clicked()));
#endif
	QPopupMenu *advanced = new QPopupMenu(t);
	if (password != NULL) {
		advanced->insertItem(trUtf8("&Set Time"), this,
				SLOT(setTime_clicked()));
		advanced->insertItem(trUtf8("&Backup to PenDrive"), this,
				SLOT(usb_clicked()));
	}
	if (ts_calibrate) {
		advanced->insertItem(trUtf8("&Calibrate Touchscreen"), this,
				SLOT(calibrate_clicked()));
	}
	if (advanced->count())
		m->insertItem(trUtf8("&Advanced"), advanced);

	QPopupMenu *help = new QPopupMenu(t);
	help->insertSeparator();
	help->insertItem(trUtf8("&About"), this, SLOT(about()));


#if QT_VERSION > 0x30000
	if (swap_cursors_style == 2) {
		const QImage *arrow_r = load_image(arrow_red_trans);
		m->insertItem(QPixmap(*arrow_r), this, SLOT(swap_cursor0()));

		const QImage *hnd = load_image(hand);
		m->insertItem(QPixmap(*hnd), this, SLOT(swap_cursor2()));

		const QImage *arrow_g = load_image(arrow_green_trans);
		m->insertItem(QPixmap(*arrow_g), this, SLOT(swap_cursor1()));
	}
#endif
	
	m->insertSeparator();
	m->insertItem(trUtf8("&Help"), help);

#ifdef AR3CMON_MENU_FONT_SIZE
	QFont *fnt = new QFont();
	fnt->setPointSize(AR3CMON_MENU_FONT_SIZE);
	m->setFont(*fnt);
	file->setFont(*fnt);
	mode->setFont(*fnt);
	tr_menu->setFont(*fnt);
	advanced->setFont(*fnt);
	help->setFont(*fnt);
	if (swap_cursors_style == 1) {
		options->setFont(*fnt);
	}
#endif


	t->setMinimumHeight(m->height());
	RootLayout->addWidget(t);
	t->show();
}

void Ar3cMonitor::createStatusLabel(QString text)
{
	QPushButton *b;

	if (StatusLabel) {
		StatusLabel->setText(text);
		return;
	}
	StatusLabel = new QLabel(text, BodyWidget);
	BodyLayout->insertWidget(0, StatusLabel);
	StatusLabel->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
	StatusLabel->show();

	QVBoxLayout *l = new QVBoxLayout(StatusLabel);
	l->addStretch(100);

	if (show_changesvr) {
		b = new QPushButton(trUtf8("Change Server"), StatusLabel);
		connect(b, SIGNAL(clicked()), this, SLOT(set_hostname()));
		l->addWidget(b);
	}
	
	b = new QPushButton(tr("Review"), StatusLabel);
	b->setEnabled(TheCommObject->canGoOffline());
	connect(b, SIGNAL(clicked()), TheCommObject, SLOT(goOffline()));
	l->addWidget(b);
	b->show();
}

void Ar3cMonitor::setTitle(QString title)
{
#ifndef TIME_IN_STATUS_BAR
	TitleLabel->setText(title);
#endif
}

void Ar3cMonitor::jump_clicked()
{
	JumpDialog *jd = new JumpDialog(CO->cursorStamp(), this);

	jd->exec();
}

void Ar3cMonitor::calibrate_clicked()
{
#ifndef __WIN32__
	size_t len = strlen(ts_calibrate);
	char *command;
	int do_fork = 0;

	if (ts_calibrate[len - 1] == '&')
		do_fork = 1;

	if (do_fork && fork())
		return;

	command = strdup(ts_calibrate);
	if (!command)
		_exit(EXIT_FAILURE);
	if (do_fork)
		command[len - 1] = '\0';
	system(command);
	_exit(EXIT_FAILURE);
#endif
}

void Ar3cMonitor::setTime_clicked()
{
#ifndef NO_SETTIME
	SetTime *st = new SetTime(CO->cursorStamp(), this);
	st->setCaption(trUtf8("Adjust datalogger time"));

	st->exec();
#endif
}

void Ar3cMonitor::usb_clicked()
{
#ifndef NO_PENDRIVE
	PenDrive *pd = new PenDrive(CO->cursorStamp(), this);
	pd->setCaption(trUtf8("Back up data to pendrive"));

	pd->exec();
#endif
}

void Ar3cMonitor::about()
{
	AboutDialog *about = new AboutDialog(this);

	about->exec();
}

void Ar3cMonitor::setupTitleBar()
{
	QWidget *w = new QWidget(RootWidget);
	Layout->addWidget(w);
	w->show();

	QHBoxLayout *hl = new QHBoxLayout(w, 8);
	TitleLabel = new QLabel(w, tr("Connecting..."));
	TitleLabel->setAlignment(Qt::AlignCenter);

	QFont titfont(font());
	titfont.setBold(true);
	titfont.setPointSize(titfont.pointSize() * 4 / 3);

	TitleLabel->setFont(titfont);
	hl->addWidget(TitleLabel, 100);
	TitleLabel->show();

	ClockLabel = new QLabel(w);
	ClockLabel->setAlignment(Qt::AlignCenter);
	QFont clockfont(font());
	clockfont.setPointSize(clockfont.pointSize() * 4 / 3);
	ClockLabel->setFont(clockfont);
	hl->addWidget(ClockLabel);
	ClockLabel->show();

	if (!show_menu) {
		jmp_button = new QPushButton(tr("jump"), w);
		connect(jmp_button, SIGNAL(clicked()), this,
				SLOT(jump_clicked()));
		hl->addWidget(jmp_button);
		jmp_button->show();
	}
}

void Ar3cMonitor::newTimestamp()
{
#ifndef TIME_IN_STATUS_BAR
	if (!ClockLabel)
		return;
#endif
	unsigned long long stamp = CO->cursorStamp();
	time_t tim = stamp / 1000000000ULL;
	struct tm *tm = localtime(&tim);

	if (!tm)
		return;
	char t[128];
#ifndef TIME_IN_STATUS_BAR
	strftime(t, sizeof(t), (CO->reviewMode() || CO->offlineMode()) ?
		"* %d %b %Y  %H:%M:%S *" : "  %H:%M:%S  ", tm);
	ClockLabel->setText(QString::fromLocal8Bit(t));
#endif
	strftime(t, sizeof(t), "%d %b %Y  %H:%M:%S", tm);
	StatusBar->newTimestamp(QString::fromLocal8Bit(t));
}
	
extern int screensaver_enabled;

void Ar3cMonitor::interface_reinit()
{
	sigset_t sigs;

	if (RootWidget != NULL)
		delete RootWidget;

#ifndef __WIN32__
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGUSR1);
	sigaddset(&sigs, SIGUSR2);

	sigprocmask(SIG_BLOCK, &sigs, NULL);
	if (SS != NULL)
		delete SS;

	SS = new QWidget();
	if (screensaver_enabled) {
		SS->showFullScreen();
		SS->show();
	}
	sigprocmask(SIG_UNBLOCK, &sigs, NULL);
#else
	if (SS != NULL)
		delete SS;

	SS = new QWidget();
#endif

	RootWidget = new QWidget(BodyWidget);
	BodyLayout->insertWidget(0, RootWidget);
	RootWidget->hide();

	CurrentStack = NULL;
	CurrentHiddenTab = NULL;
	CurrentDzwonek = NULL;
	CurrentAnalogGraph = NULL;
	AttrWidget = NULL;

	Layout = new QVBoxLayout(RootWidget);

#ifndef TIME_IN_STATUS_BAR
	setupTitleBar();
#endif
	HLayout = new QHBoxLayout(Layout);

	VLayout = new QVBoxLayout(HLayout);
	HLayout->setStretchFactor(VLayout, 1000);

	CurrentWidget = RootWidget;
	CurrentLayout = VLayout;

	Tabs = new MyTabWidget(RootWidget);
	VLayout->addWidget(Tabs);
	Tabs->show();
}

void Ar3cMonitor::commConnected()
{
	qWarning("connected");
	createStatusLabel(tr("Connected, user interface initialization "
				"in progress..."));
	interface_reinit();

	if (scheme >= 0) {
		if (conf_locale)
			TheCommObject->request_locale(conf_locale);

		if (led_no_symbol)
			TheCommObject->set_opt(AR3CM_PROTO_OPT_LED_NO_SYMBOL, 1);

		TheCommObject->requestScheme(scheme);

		TheCommObject->get_token();
	}
}

void Ar3cMonitor::commOfflineConnected()
{
	commConnected();
	CO->connected();
}

void Ar3cMonitor::commConnecting()
{
	Q_CHECK_PTR(StatusLabel);
	createStatusLabel(tr("Connecting to datalogger..."));
}

void Ar3cMonitor::commError()
{
	delete RootWidget;
	RootWidget = NULL;
	ClockLabel = NULL;
	TabsCnt = 0;
	for (int i = 0; i < AR3CMON_MAX_TABS; i++)
		TabsTbl[i] = NULL;
	createStatusLabel(tr("Communication Error"));
}

#if QT_VERSION < 0x40100
class MinWidget : public QWidget {
public:
	MinWidget(QWidget *parent = 0, const char *name = 0, Qt::WFlags f = 0)
		: QWidget(parent, name, f)
	{
	}

	QSize sizeHint(void) const
	{
		return minimumSizeHint();
	}
};
#endif

void Ar3cMonitor::newTab(QString caption, bool hidden)
{
#if QT_VERSION >= 0x40100
	QScrollArea *scroll = new QScrollArea();
	CurrentWidget = new QWidget();

	scroll->setWidgetResizable(true);
	scroll->setWidget(CurrentWidget);
	CurrentWidget->setBackgroundColor(getBgColor());
	CurrentWidget->setAutoFillBackground(true);
#else
	QScrollView *scroll = new QScrollView();
	CurrentWidget = new MinWidget(scroll);

	scroll->setResizePolicy(QScrollView::AutoOneFit);
	scroll->addChild(CurrentWidget);

	CurrentWidget->setBackgroundColor(getBgColor());
#endif

	CurrentDzwonek = NULL;
	CurrentLayout = new QVBoxLayout(CurrentWidget, 0);
	if (TabsCnt >= AR3CMON_MAX_TABS) {
		qWarning("Tabs limit exceeded, new tab will not be visible");
		return;
	} else {
		TabsTbl[TabsCnt] = scroll;
		TabsCaptions[TabsCnt++] = caption;
	}
	if (touchscreen) {
		Tabs->addTab(scroll, "  " + caption + "  ");
	} else {
		Tabs->addTab(scroll, caption);
	}
#if QT_VERSION >= 0x30000
	if (hidden)
		Tabs->removePage(scroll);
#endif
}

void Ar3cMonitor::hideHiddenTab()
{
	QWidget *t = CurrentHiddenTab;
	if (!t)
		return;
	CurrentHiddenTab = NULL;
	Tabs->removePage(t);
}

void Ar3cMonitor::activateTab(int index)
{
	if (index < 0 || index >= TabsCnt || !TabsTbl[index]) {
		qWarning("invalid tab index while calling activateTab()");
		return;
	}

	QWidget *w = TabsTbl[index];
	hideHiddenTab();
#if QT_VERSION >= 0x30000
	if (Tabs->indexOf(w) == -1) {
		Tabs->addTab(w, TabsCaptions[index]);
		Tabs->showPage(w);
		w->show();
		CurrentHiddenTab = w;
	} else
#endif
		Tabs->showPage(w);
}

void Ar3cMonitor::newLeftbox()
{
	CurrentDzwonek = NULL;
	CurrentLayout = new QVBoxLayout(RootWidget);
	HLayout->insertLayout(0, CurrentLayout, 0);
	HLayout->setStretchFactor(CurrentLayout, 0);
	CurrentWidget = RootWidget;
}

void Ar3cMonitor::newScreenSaver()
{
	CurrentDzwonek = NULL;
	CurrentLayout = new QVBoxLayout(SS, 2);
	CurrentWidget = SS;
}

void Ar3cMonitor::newTClockLabel(const char *format)
{
	TClockLabel *cl = new TClockLabel(format, CurrentWidget);
	cl->setBackgroundColor(getBgColor());
	QPalette tmp = cl->palette();
	tmp.setColor(QColorGroup::Text, getColor());
	cl->setPalette(tmp);
	AttrWidget = cl;
	CurrentLayout->addWidget(cl);
	cl->show();
}

void Ar3cMonitor::newObject(const char *format)
{
	ObjectWidget *cl = new ObjectWidget(format, CurrentWidget);
	cl->setBackgroundColor(getBgColor());
	AttrWidget = cl;
	CurrentLayout->addWidget(cl);
	cl->show();
}

void Ar3cMonitor::newBottombox()
{
	QWidget *w = new QWidget(RootWidget);
	w->setSizePolicy(QSizePolicy(QSizePolicy::Preferred,
			QSizePolicy::Maximum));
	w->setBackgroundColor(getBgColor());
#if QT_VERSION >= 0x40100
	w->setAutoFillBackground(true);
#endif
	VLayout->addWidget(w, 0);
	CurrentDzwonek = NULL;
	CurrentLayout = new QVBoxLayout(w, 2);
	CurrentWidget = w;
	w->show();
}

void Ar3cMonitor::newLabel(QString text)
{
	QLabel *l = new QLabel(text, CurrentWidget);
	l->setBackgroundColor(getBgColor());
#if QT_VERSION >= 0x40100
	l->setAutoFillBackground(true);
#endif
	QPalette tmp = l->palette();
	tmp.setColor(QColorGroup::Foreground, getColor());
	tmp.setColor(QColorGroup::Text, getColor());
	l->setPalette(tmp);

	AttrWidget = l;
	CurrentLayout->addWidget(l);
	l->show();
}

void Ar3cMonitor::newEmpty()
{
	QWidget *w = new QWidget(CurrentWidget);
	w->setBackgroundColor(getBgColor());
#if QT_VERSION >= 0x40100
	w->setAutoFillBackground(true);
#endif
	AttrWidget = w;
	CurrentLayout->addWidget(w);
	w->show();
}



void Ar3cMonitor::newBigLabel(QString text)
{
	QLabel *l = new QLabel(CurrentWidget);
	l->setSizePolicy(QSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed));
	l->setBackgroundColor(getBgColor());
	l->setTextFormat(Qt::RichText);
	l->setText(QString("<center><font size=\"+2\">"
			"<b>%1</b></font></center>").arg(text));
	QPalette tmp = l->palette();
	tmp.setColor(QColorGroup::Text, getColor());
	l->setPalette(tmp);

	AttrWidget = l;
	CurrentLayout->addWidget(l);
	l->show();
}



void Ar3cMonitor::currentPush()
{
	QColor tmp;
	tmp.setRgb(0, 0, 0);
	LayoutsStack *s = new LayoutsStack();
	s->bgcolor = tmp;
	tmp.setRgb(255, 255, 255);
	s->color = tmp;
	s->Next = CurrentStack;
	s->CurrentWidget = CurrentWidget;
	s->CurrentLayout = CurrentLayout;
	CurrentStack = s;
}

void Ar3cMonitor::newHBox()
{
	currentPush();
	CurrentLayout = new QHBoxLayout(CurrentLayout, 0);
	CurrentLayout->setMargin(6);
}

void Ar3cMonitor::newVBox()
{
	currentPush();
	CurrentLayout = new QVBoxLayout(CurrentLayout, 0);
	CurrentLayout->setMargin(6);
}

void Ar3cMonitor::newHLayout()
{
	currentPush();
	CurrentLayout = new QHBoxLayout(CurrentLayout, 0);
}

void Ar3cMonitor::newVLayout()
{
	currentPush();
	CurrentLayout = new QVBoxLayout(CurrentLayout, 0);
}

void Ar3cMonitor::endLayout()
{
	if (CurrentStack == NULL) {
		qWarning("layout stack overflow, please check configuration");
		return;
	}
	CurrentLayout = CurrentStack->CurrentLayout;
	CurrentWidget = CurrentStack->CurrentWidget;
	CurrentStack = CurrentStack->Next;
	AttrWidget = NULL;
}

void Ar3cMonitor::setBgColor(QColor bgcolor)
{
	if (CurrentStack != NULL) {
		CurrentStack->bgcolor = bgcolor;
	} else {
		QPalette tmp = palette();
		tmp.setColor(QColorGroup::Background, bgcolor);
		setPalette(tmp);
	}
}

void Ar3cMonitor::setColor(QColor color)
{
	if (CurrentStack != NULL) {
		CurrentStack->color = color;
	} else {
		QPalette tmp = palette();
		tmp.setColor(QColorGroup::Foreground, color);
		setPalette(tmp);
	}
}

QColor Ar3cMonitor::getBgColor() const
{
	return (CurrentStack != NULL) ? CurrentStack->bgcolor : Qt::black;
}

QColor Ar3cMonitor::getColor() const
{
	return (CurrentStack != NULL) ? CurrentStack->color : Qt::white;
}

void Ar3cMonitor::newSeparator()
{
	bool horiz = CurrentLayout->direction() == QBoxLayout::TopToBottom;
	MySeparator *w = new MySeparator(horiz, CurrentWidget);

	if (horiz) {
		w->setMaximumHeight(7);
		w->setMinimumHeight(7);
	} else {
		w->setMaximumWidth(7);
		w->setMinimumWidth(7);
	}
	CurrentLayout->addWidget(w);
	w->show();
}

void Ar3cMonitor::newSpacer()
{
	CurrentLayout->addSpacing(6);
}

void Ar3cMonitor::newStretch()
{
	CurrentLayout->addStretch(1000);
}

void Ar3cMonitor::addLed(int physical, QString caption, QColor ledcolor,
		QColor activetextcolor, QColor ledinactivecolor, bool diag)
{
	QWidget *w = new QWidget(CurrentWidget);

	CurrentLayout->addWidget(w, 0);

	w->setBackgroundColor(Qt::black);
#if QT_VERSION >= 0x40100
	w->setAutoFillBackground(true);
#endif

	w->show();

	QHBoxLayout *l = new QHBoxLayout(w);

	LedWidget *led = new LedWidget(physical, ledcolor, ledinactivecolor, w);
	l->addWidget(led, 0, Qt::AlignLeft | Qt::AlignVCenter);
	led->show();
	LedLabel *lbl = new LedLabel(physical, caption, activetextcolor, w);
	l->addWidget(lbl, 100, Qt::AlignLeft | Qt::AlignVCenter);
	lbl->show();

	AttrWidget = w;

	if (physical == 0xffff)
		return;

#if defined(__WIN32__)
	connect(CO, SIGNAL(sigUpdateTimer()), led, SLOT(update()));
	connect(CO, SIGNAL(sigUpdateTimer()), lbl, SLOT(update()));
#else	
	connect(CO->bobject(physical), SIGNAL(change()),
			led, SLOT(update()));
	connect(CO->bobject(physical), SIGNAL(change()),
			lbl, SLOT(update()));
#endif
	connect(CO, SIGNAL(sigPleaseUpdate()), led, SLOT(update()));
	connect(CO, SIGNAL(sigPleaseUpdate()), lbl, SLOT(update()));

	if (diag)
		CO->setAlwaysDefined(physical);
	else {
		CO->clearAlwaysDefined(physical);
		connect(CO, SIGNAL(sigHaveDataChanged()), led,
				SLOT(haveDataChanged()));
		led->haveDataChanged();
	}
}

void Ar3cMonitor::newGraph()
{
	CurrentAnalogGraph = new AnalogGraph(CurrentWidget);
	CurrentLayout->addWidget(CurrentAnalogGraph, 100);
	connect(CO, SIGNAL(sigUpdateTimer()), CurrentAnalogGraph,
			SLOT(update_timer()));
	CurrentAnalogGraph->show();
}

void Ar3cMonitor::newTimeScale()
{
	TimeScale *ts = new TimeScale(CurrentWidget);

	CurrentLayout->addWidget(ts, 100);
	connect(CO, SIGNAL(sigUpdateTimer()), ts, SLOT(update()));
	ts->show();
}

void Ar3cMonitor::setSizeHint(int hint)
{
	if (CurrentAnalogGraph == NULL) {
		qWarning("size hint is only supported for AnalogGraph");
		return;
	}
	CurrentAnalogGraph->addSizeHint(hint);
}

void Ar3cMonitor::setMinHeightAttr(int hint)
{
	if (AttrWidget == NULL) {
		qWarning("misplaced MinHeight");
		return;
	}
	AttrWidget->setMinimumHeight(hint);
}

void Ar3cMonitor::setMaxHeightAttr(int hint)
{
	if (AttrWidget == NULL) {
		qWarning("misplaced MaxHeight");
		return;
	}
	AttrWidget->setMaximumHeight(hint);
}

void Ar3cMonitor::setMinWidthAttr(int hint)
{
	if (AttrWidget == NULL) {
		qWarning("misplaced MinWidth");
		return;
	}
	AttrWidget->setMinimumWidth(hint);
}

void Ar3cMonitor::setMaxWidthAttr(int hint)
{
	if (AttrWidget == NULL) {
		qWarning("misplaced MaxWidth");
		return;
	}
	AttrWidget->setMaximumWidth(hint);
}

void Ar3cMonitor::setStretchAttr(int hint)
{
	if (AttrWidget == NULL) {
		qWarning("misplaced stretch attribute");
		return;
	}
	CurrentLayout->setStretchFactor(AttrWidget, hint);
}

void Ar3cMonitor::hide_cond(int physical, int not_cond)
{
	if (AttrWidget == NULL) {
		qWarning("misplaced stretch attribute");
		return;
	}

	HideWidget *hide = new HideWidget(physical, not_cond, AttrWidget,
			CurrentWidget);
	CurrentLayout->addWidget(hide);

#if defined(__WIN32__)
	connect(CO, SIGNAL(sigUpdateTimer()), hide, SLOT(update_data()));
#else
	connect(CO->bobject(physical), SIGNAL(change()),
			hide, SLOT(update_data()));
#endif
	connect(CO, SIGNAL(sigPleaseUpdate()), hide, SLOT(update_data()));
}

void Ar3cMonitor::addPlot(int physical, QString caption, QColor color,
	float min, float max, float uv, QString unit)
{
	CurrentAnalogGraph->addPlot(physical, caption, color, min, max, uv,
			unit, CO->get_prec());
}

void Ar3cMonitor::addHGrid(int nlines, int *lines, QColor color)
{
	CurrentAnalogGraph->addHGrid(nlines, lines, color);
}

void Ar3cMonitor::newDzwonek(int physical, QString symbol, QColor activecolor,
	QColor inactivecolor)
{
	if (physical == 0xffff)
		return;

	DzwonekGraph *dg = CurrentDzwonek;

	if (dg == NULL) {
		dg = new DzwonekGraph(CurrentWidget);
		CurrentLayout->addWidget(dg);
		AttrWidget = dg;
		CurrentDzwonek = dg;
		connect(CO, SIGNAL(sigUpdateTimer()), dg, SLOT(update()));
		dg->show();
	}
	dg->addSignal(physical, activecolor, inactivecolor, symbol);
}

void Ar3cMonitor::newBinGraph(int physical, QString symbol, QColor activecolor,
		QColor inactivecolor)
{
	if (physical == 0xffff)
		return;

	BinaryGraph *g = new BinaryGraph(physical, activecolor, inactivecolor,
			symbol, CurrentWidget);
	AttrWidget = g;
	CurrentLayout->addWidget(g);
	connect(CO, SIGNAL(sigUpdateTimer()), g, SLOT(update()));
	g->show();
}

void Ar3cMonitor::newVMeter(int physical, QString, QColor color,
	float min, float max, float uv, unsigned int flags)
{
	MeterVWidget *w = new MeterVWidget(physical, color, max, min, uv,
			flags, CO->get_prec(), CurrentWidget);
	AttrWidget = w;
	CurrentLayout->addWidget(w, 1000);
	connect(CO, SIGNAL(sigUpdateTimer()), w, SLOT(updateMeter()));
	w->show();
}

void Ar3cMonitor::newAMeter(int physical, QString caption, QColor color,
	float min, float max, float uv, int angle, int scale,
	QString unit, unsigned long flags)
{
	MeterAWidget *w = new MeterAWidget(physical, color, max, min, uv,
			angle, scale, caption, unit, flags, CO->get_prec(),
			CurrentWidget);
	AttrWidget = w;
	CurrentLayout->addWidget(w, 1000);
	connect(CO, SIGNAL(sigUpdateTimer()), w, SLOT(updateMeter()));
	w->show();
}

void Ar3cMonitor::newDMeter(int physical, QColor color, float uv,
			QString unit, unsigned long flags,
			QString prefix)
{
	MeterDWidget *w = new MeterDWidget(physical, color, uv, unit, flags,
			prefix, CurrentWidget);

	w->setBackgroundColor(getBgColor());
	QPalette tmp = w->palette();
	tmp.setColor(QColorGroup::Text, getColor());
	w->setPalette(tmp);

	AttrWidget = w;
	CurrentLayout->addWidget(w, 1000);
	CurrentLayout->setStretchFactor(w, 0);
	connect(CO, SIGNAL(sigUpdateTimer()), w, SLOT(updateMeter()));
	w->show();
}

void Ar3cMonitor::newButton(QString caption, Action * action)
{
	QPushButton *b = new QPushButton(CurrentWidget);
//	b->setBackgroundColor(getBgColor());
	b->setText(caption);
	QPalette tmp = b->palette();
	tmp.setColor(QColorGroup::Text, getColor());
	b->setPalette(tmp);

	b->setSizePolicy(QSizePolicy(QSizePolicy::Minimum,
				QSizePolicy::Minimum));
	connect(b, SIGNAL(clicked()), action, SLOT(run()));
	AttrWidget = b;
	CurrentLayout->addWidget(b);
	b->show();
}

void Ar3cMonitor::newSzybowskaz(int posPhys, int velPhys, float posUV,
	float velUV, float posMin, float posMax, float velMin,
	float velMax, int level_count, struct ar3cm_level *levels,
	int left)
{
	Wskaz *w = new Wskaz(posPhys, velPhys, posUV, velUV, posMin, posMax,
			velMin, velMax, level_count, levels, left,
			CurrentWidget);
	AttrWidget = w;
	CurrentLayout->addWidget(w);
	w->show();
}

void Ar3cMonitor::visSetupEnd()
{
	if (CurrentStack != NULL)
		qWarning("layout stack underflow, please check configuration");

	delete StatusLabel;
	StatusLabel = NULL;
	RootWidget->show();

	Tabs->setCurrentPage(tabIndex);

	connect(Tabs, SIGNAL(currentChanged(QWidget *)), SLOT(tabChanged()));
}

void Ar3cMonitor::tabChanged()
{
	tabIndex = Tabs->currentPageIndex();
	Q_ASSERT(Tabs->currentPage() != CurrentHiddenTab);
	hideHiddenTab();
}

void Ar3cMonitor::increase_timerange()
{
	if (CO->timeRange() >= 3 * 60 * 60 * 1000000000ULL)
		return;
	if (CO->reviewMode())
		CO->advanceCursor(CO->timeRange() / 4);
	CO->setTimeRange(CO->timeRange() * 3 / 2);
}

void Ar3cMonitor::timerange_15s()
{
	CO->setTimeRange(15 * 1000000000ULL);
}

void Ar3cMonitor::timerange_30s()
{
	CO->setTimeRange(30 * 1000000000ULL);
}

void Ar3cMonitor::timerange_1m()
{
	CO->setTimeRange(60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_2m()
{
	CO->setTimeRange(2 * 60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_5m()
{
	CO->setTimeRange(5 * 60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_10m()
{
	CO->setTimeRange(10 * 60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_15m()
{
	CO->setTimeRange(15 * 60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_30m()
{
	CO->setTimeRange(30 * 60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_1h()
{
	CO->setTimeRange(1 * 60 * 60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_2h()
{
	CO->setTimeRange(2 * 60 * 60 * 1000000000ULL);
}

void Ar3cMonitor::timerange_3h()
{
	CO->setTimeRange(3 * 60 * 60 * 1000000000ULL);
}

void Ar3cMonitor::select_active_layer()
{
	CO->set_layer(-1);
}

void Ar3cMonitor::select_0_layer()
{
	CO->set_layer(0);
}

void Ar3cMonitor::select_1_layer()
{
	CO->set_layer(1);
}

void Ar3cMonitor::select_2_layer()
{
	CO->set_layer(2);
}

void Ar3cMonitor::select_3_layer()
{
	CO->set_layer(3);
}

void Ar3cMonitor::select_4_layer()
{
	CO->set_layer(4);
}

void Ar3cMonitor::select_5_layer()
{
	CO->set_layer(5);
}

void Ar3cMonitor::select_6_layer()
{
	CO->set_layer(6);
}

void Ar3cMonitor::select_7_layer()
{
	CO->set_layer(7);
}

void Ar3cMonitor::decrease_timerange()
{
	if (CO->timeRange() <= 1000000000ULL)
		return;
	CO->setTimeRange(CO->timeRange() * 2 / 3);
	if (CO->reviewMode())
		CO->advanceCursor(-(long long) CO->timeRange() / 4);
}

void Ar3cMonitor::handle_key(QKeyEvent *e)
{
	switch (e->key()) {
	case Qt::Key_Q:
		if (key_q_action && !fork()) {
			execl(key_q_action, key_q_action, NULL);
			_exit(EXIT_FAILURE);
		}
		break;
	case Qt::Key_X:
		if (key_x_action && !fork()) {
			execl(key_x_action, key_x_action, NULL);
			_exit(EXIT_FAILURE);
		}
		break;
	case Qt::Key_Left:
		if (CO->reviewMode())
			CO->advanceCursor(-(long long) CO->timeRange() / 10);
		else if (TabsCnt)
			Tabs->setCurrentPage(tabIndex - 1);
		break;
	case Qt::Key_Right:
		if (CO->reviewMode())
			CO->advanceCursor(CO->timeRange() / 10);
		else if (TabsCnt)
			Tabs->setCurrentPage(tabIndex + 1);
		break;
	case Qt::Key_Up:
		increase_timerange();
		break;
	case Qt::Key_Down:
		decrease_timerange();
		break;
	case Qt::Key_Return:
	case Qt::Key_Enter:
		if (TabsCnt)
			Tabs->setCurrentPage(0);
		break;
#ifdef __SIMPAD__
	case Qt::Key_F34:	/* poweroff */
		system("halt -f -p");
		break;
	case Qt::Key_Escape:
		qa->quit();
		return;
#endif
	default:
		//QMessageBox::information(this, QString("AR-3c Monitor"), QString("Unbound key pressed: %1").arg((int)e->key()));
		e->ignore();
		return;
	}
	e->accept();
}

void Ar3cMonitor::keyPressEvent(QKeyEvent * e)
{
	this->handle_key(e);
}

void Ar3cMonitor::menu_quit()
{
	qa->quit();
}

void Ar3cMonitor::closeEvent(QCloseEvent *)
{
	qa->quit();
}

void Ar3cMonitor::swap_cursor0()
{
	if (options) {
		options->setItemChecked(swapcur0_id, true);
		options->setItemChecked(swapcur1_id, false);
		options->setItemChecked(swapcur2_id, false);
	}
	CO->cursor_mode_set(CURSOR_MODE_LRM);
}

void Ar3cMonitor::swap_cursor1()
{
	if (options) {
		options->setItemChecked(swapcur0_id, false);
		options->setItemChecked(swapcur1_id, true);
		options->setItemChecked(swapcur2_id, false);
	}
	CO->cursor_mode_set(CURSOR_MODE_RLM);
}

void Ar3cMonitor::swap_cursor2()
{
	if (options) {
		options->setItemChecked(swapcur0_id, false);
		options->setItemChecked(swapcur1_id, false);
		options->setItemChecked(swapcur2_id, true);
	}
	CO->cursor_mode_set(CURSOR_MODE_MLR);
}

QString text2unicode(QString a)
{
	QTextCodec *codec = QTextCodec::codecForName(config_encoding);
        return (codec != NULL) ? codec->toUnicode(a) : a;
}

void Ar3cMonitor::set_hostname()
{
#ifndef QT_NO_INPUTDIALOG
	bool ok;
	QString text = QInputDialog::getText("ar3cmonitor",
			trUtf8("Select hostname:"), QLineEdit::Normal,
			QString(ar3cmon_host), &ok, this);

	if (ok && !text.isEmpty()) {
		char *t;

		t = strdup(text.ascii());
		if (t) {
			free((char *)ar3cmon_host);
			ar3cmon_host = t;
			TheCommObject->errorHandler(-8);
		}
	}
#endif
}
