//
// C++ Interface: commobj
//
// Description: 
//
//
// Author: Szymon Acedanski <accek@podlesie.net>, (C) 2004
// Copyright (C) 2006..2008  Krzysztof Mazur <krzysiek@podlesie.net>
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <stdio.h>
#include <limits.h>

#ifdef UNIX
# include <unistd.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <signature.h>

#ifndef __WIN32__
#include <sys/mman.h>
#endif

#ifdef HAVE_SSL
#include <openssl/sha.h>
#endif

#include "commobj.h"
#include "ar3cmonitor.h"
#include "ar3cm_proto.h"
#include "consts.h"
#include "action.h"
#include "control.h"
#include "analogcache.h"
#include <fonts.h>

const char *ar3cmon_host;
unsigned short ar3cmon_port = AR3CMON_DFLT_PORT;
extern char *config_encoding;

CommObject *TheCommObject;

CommObject::CommObject(Ar3cMonitor *owner)
	: QObject(owner), Owner(owner)
{
	lastTimestamp = 0;
	PacketHdr.type = cpu_to_amp16(0);
	offline = false;
	token_state = TOKEN_NONE;
	FullyConnected = false;
	SetupSeq = new unsigned char[AR3CMON_SETUPSEQ_BUF];
	SetupSeqOff = -1;
	orientation = 0;
	Socket = new QSocket(this);
	connect(Socket, SIGNAL(connected()), SLOT(connectedHandler()));
	connect(Socket, SIGNAL(connectionClosed()), SLOT(closedHandler()));
	connect(Socket, SIGNAL(delayedCloseFinished()), SLOT(closedHandler()));
	connect(Socket, SIGNAL(error(int)), SLOT(errorHandler(int)));
	connect(Socket, SIGNAL(readyRead()), SLOT(readHandler()));

	Watchdog = new QTimer(this);
	connect(Watchdog, SIGNAL(timeout()), SLOT(watchdogHandler()));

	FullRequestTimer = new QTimer(this);
	connect(FullRequestTimer, SIGNAL(timeout()), SLOT(requestData()));

	reconnect_timer = new QTimer(this);
	connect(reconnect_timer, SIGNAL(timeout()), SLOT(connectToAr3c()));

	password = NULL;
	password_size = 0;
}

CommObject::~CommObject()
{
	delete []SetupSeq;
}

void CommObject::connectToAr3c()
{
	if (Socket->state() != QSocket::Idle) {
		errorHandler(-6);
		return;
	}

	if (offline) {
		offline = false;
		qWarning("leaving offline mode");
		emit sigError();
	}

	Socket->connectToHost(ar3cmon_host, ar3cmon_port);
	Watchdog->start(AR3CMON_CONNECTION_TIMEOUT, TRUE);
	emit sigConnecting();
}

void CommObject::errorHandler(int err)
{
	qWarning("Error %d occured", err);
	FullyConnected = false;
	CO->noData();
	Watchdog->start(AR3CMON_CONNECTION_TIMEOUT, TRUE);
	if (Socket->state() == QSocket::Closing) {
		qWarning("socket error while QSocket::Closing!");
		closedHandler();
		emit sigError();
		return;
	}
#if QT_VERSION >= 0x30000
	Socket->clearPendingData();
	Socket->close();
#else
	delete Socket;
	Socket = new QSocket(this);
	Q_CHECK_PTR(Socket);
	connect(Socket, SIGNAL(connected()), SLOT(connectedHandler()));
	connect(Socket, SIGNAL(connectionClosed()), SLOT(closedHandler()));
	connect(Socket, SIGNAL(delayedCloseFinished()), SLOT(closedHandler()));
	connect(Socket, SIGNAL(error(int)), SLOT(errorHandler(int)));
	connect(Socket, SIGNAL(readyRead()), SLOT(readHandler()));
#endif
	emit sigError();
	if (Socket->state() == QSocket::Idle)
		closedHandler();
}

void CommObject::goOnline()
{
	if (offline)
		connectToAr3c();
}

void CommObject::goOffline()
{
	if (offline)
		return;
	if (!canGoOffline()) {
		qWarning("offline mode requested, although preconditions not "
				"met");
		return;
	}
	offline = true;
	Watchdog->stop();
#if QT_VERSION >= 0x30000
	Socket->clearPendingData();
	Socket->close();
#else
	delete Socket;
	Socket = new QSocket(this);
	Q_CHECK_PTR(Socket);
	connect(Socket, SIGNAL(connected()), SLOT(connectedHandler()));
	connect(Socket, SIGNAL(connectionClosed()), SLOT(closedHandler()));
	connect(Socket, SIGNAL(delayedCloseFinished()), SLOT(closedHandler()));
	connect(Socket, SIGNAL(error(int)), SLOT(errorHandler(int)));
	connect(Socket, SIGNAL(readyRead()), SLOT(readHandler()));
#endif
	emit sigError();
	if (Socket->state() == QSocket::Idle)
		closedHandler();
}

bool CommObject::reallyConnected()
{
	return FullyConnected && Socket->state() == QSOCK_CONNECTED;
}

bool CommObject::canGoOffline()
{
	return (FullyConnected || Socket->state() != QSOCK_CONNECTED)
		&& SetupSeqOff > 0;
}

void CommObject::closedHandler()
{
	unsigned long long ts;
	
	Q_ASSERT(Socket->state() == QSocket::Idle);
	emit sigError();

	ts = CO->timeStamp();

	CO->reset();

	if (ts) {
		lastTimestamp = CO->beginUndefinedRegion(0);
	}

	/* WHAT'S THIS? */
	if (offline) {
		startOfflineProtocol();
		return;
	}
	reconnect_timer->start(300, TRUE);
}

void CommObject::connectedHandler()
{
	PacketHdr.type = cpu_to_amp16(0);
	SetupSeqOff = 0;
	CO->noData();
	CO->nowCanDownload();
	emit sigReallyConnected();
	emit sigConnected();
}

void CommObject::sendSimpleCommand(int opcode)
{
	struct ar3cm_proto_hdr hdr;
	
	hdr.type = cpu_to_amp16(opcode);
	hdr.size = cpu_to_amp16(0);
	Q_LONG w = Socket->writeBlock((const char *) &hdr, sizeof(hdr));
	if (w != sizeof(hdr))
		qWarning("error sending simple command %d", opcode);
	Socket->flush();
}

void CommObject::sendCommand(int opcode, const void *buf, int len)
{
	struct ar3cm_proto_hdr hdr;

	if ((Socket->state() == QSocket::Closing) || offline)
		return;
	
	hdr.type = cpu_to_amp16(opcode);
	hdr.size = cpu_to_amp16(len);
	Q_LONG w = Socket->writeBlock((const char *) &hdr, sizeof(hdr));
	if (w != sizeof(hdr))
		qWarning("error sending command header (opcode=%d)", opcode);
	w = Socket->writeBlock((const char *) buf, len);
	if (w != len)
		qWarning("error sending command data (opcode=%d)", opcode);
	Socket->flush();
}

void CommObject::readHandler()
{
	int i = 0;
	
	for (;;) {
		if (amp16_to_cpu(PacketHdr.type) == 0) {
			if ((size_t) Socket->bytesAvailable()
					< sizeof(PacketHdr))
				break;
			Socket->readBlock((char *) &PacketHdr,
				sizeof(PacketHdr));
			if (amp16_to_cpu(PacketHdr.size)
					> AR3CMON_COMM_BUF_SIZE) {
				errorHandler(-3);
				break;
			}
		}
		if (Socket->bytesAvailable() < (unsigned int)
				amp16_to_cpu(PacketHdr.size))
			break;

		unsigned char __attribute__((aligned(8)))
			buf[AR3CMON_COMM_BUF_SIZE];
		Socket->readBlock((char *) buf, amp16_to_cpu(PacketHdr.size));

		i++;
		parsePacket(&PacketHdr, buf);
		handleSetupSeq(buf);
		//fprintf(stderr, "packet %d\n", PacketHdr.type);

		PacketHdr.type = cpu_to_amp16(0);
		Watchdog->start(AR3CMON_COMMUNICATION_TIMEOUT, TRUE);
	}
}

int CommObject::check_packet_size(struct ar3cm_proto_hdr *hdr, size_t min,
		size_t max)
{
	size_t size = amp16_to_cpu(hdr->size);
	int type = amp16_to_cpu(hdr->type);

	if ((size >= min) && (size <= max))
		return 0;
	qWarning("packet with incorrect length: %02x(%d) len=%zd min=%zd "
			"max=%zd\n",
			type, type, size, min, max);
	return 1;
}

int CommObject::check_packet_size_eq(struct ar3cm_proto_hdr *hdr, size_t size)
{
	return check_packet_size(hdr, size, size);
}

int CommObject::check_packet_size_min(struct ar3cm_proto_hdr *hdr, size_t min)
{
	return check_packet_size(hdr, min, amp16_to_cpu(hdr->size));
}

QColor CommObject::hsv_qcolor(struct ar3cm_proto_hsv *c)
{
	return QColor(amp16_to_cpu(c->h), amp16_to_cpu(c->s),
			amp16_to_cpu(c->v), QColor::Hsv);
}

void CommObject::parse_binitem(struct ar3cm_proto_hdr *hdr, unsigned char *buf)
{
	struct ar3cm_proto_binitem *pkt = (struct ar3cm_proto_binitem *) buf;
	__uint8_t type;
	
	if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
		return;

	type = amp8_to_cpu(pkt->type);
	/* trim caption to not overflow */
	buf[amp16_to_cpu(hdr->size) - 1] = '\0';
	switch (type) {
	case AR3CM_BINTYPE_NORMAL:
	case AR3CM_BINTYPE_DIAG:
		Owner->addLed(amp16_to_cpu(pkt->physical),
				text2unicode(QString(pkt->caption)),
				hsv_qcolor(&pkt->ledcolor),
				hsv_qcolor(&pkt->activetextcolor),
				hsv_qcolor(&pkt->inactiveledcolor),
				type == AR3CM_BINTYPE_DIAG);
		CO->setBinaryUsed(amp16_to_cpu(pkt->physical));
		break;
	default:
		qWarning("unrecognized binary signal, ignoring...");
	}
}
			
static const char *amp_string_get(const struct amp_string *string,
		const void *base, const struct ar3cm_proto_hdr *hdr)
{
	size_t offset = amp16_to_cpu(string->offset);
	size_t size = amp16_to_cpu(hdr->size);

	if (((const char *) base)[size - 1] != '\0')
		return NULL;
	if (offset >= size)
		return NULL;
	return (const char *) base + offset;
}

void CommObject::receive_token(struct ar3cm_proto_hdr *hdr,
		struct ar3cm_proto_token *t)
{
	int ret;

	if (check_packet_size_min(hdr, sizeof(*t)))
		return;

	/*
	 * Dropping unwanted tokens make token forward attacks
	 * harder.
	 */
	if (token_state != TOKEN_WANT) {
		qWarning("unwanted token received, ignoring");
		return;
	}

	if (!ar3cm_dblock_access_ok(&t->token, amp16_to_cpu(hdr->size))) {
		qWarning("invalid token data");
		return;
	}

	ret = token_init(&token, amp32_to_cpu(t->type), amp32_to_cpu(t->ttl));
	if (ret)
		goto err;

	ret = token_load(&token, ar3cm_dblock_ptr(&t->token, t),
			ar3cm_dblock_size(&t->token));
	if (ret)
		goto err;

	token_state = TOKEN_READY;
	return;
	
err:
	qWarning("receive_token: error %d\n", ret);
	token_state = TOKEN_NONE;
}

/* buf for some types must be 8-byte aligned */
void CommObject::parsePacket(struct ar3cm_proto_hdr *hdr, unsigned char *buf)
{
	size_t size = amp16_to_cpu(hdr->size);
	unsigned int type = amp16_to_cpu(hdr->type);

	switch (type) {
	case AR3CM_PROTO_ANALOGS: {
		struct ar3cm_proto_analogs *pkt =
			(struct ar3cm_proto_analogs *) buf;
		int n;
		unsigned long long timestamp;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			return;

	       	n = (size - sizeof(struct ar3cm_proto_analogs)) / 2;

		timestamp = amp64_to_cpu(pkt->timestamp);
		if (timestamp < lastTimestamp)
			break;

		AFullRecord *r = new AFullRecord(n);
		
		for (int i = 0; i < n; i++)
			r->data[i] = ampi16_to_cpu(pkt->analogs[i]);

		if (timestamp - lastTimestamp > AR3CMON_DELTA_FOR_GAP) {
			lastTimestamp = CO->beginUndefinedRegion(0);

			/*
			 * probably some data was lost and we should
			 * ask for retransmission of full binary signal
			 * state
			 */
			sendSimpleCommand(AR3CM_PROTO_SENDBINARIES);

			if (timestamp < lastTimestamp)
				break;
		}

		if (unlikely(CO->newARecord(timestamp, r)))
			delete r;
		else if (!CO->offlineMode())
			for (int i = 0; i < n; i++)
				ACC->update_cache(i, timestamp,
						ampi16_to_cpu(pkt->analogs[i]));

		CO->ping();
		lastTimestamp = timestamp;
		break;
	}
	case AR3CM_PROTO_STOREDA: {
		struct ar3cm_proto_analogs *pkt =
			(struct ar3cm_proto_analogs *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			return;
		
		int n = (size - sizeof(struct ar3cm_proto_analogs)) / 2;
		
		AFullRecord *r = new AFullRecord(n);
		for (int i = 0; i < n; i++)
			r->data[i] = ampi16_to_cpu(pkt->analogs[i]);
		if (CO->newOffARecord(amp64_to_cpu(pkt->timestamp), r))
			delete r;
		break;
	}
	case AR3CM_PROTO_STOREDSINGLEA: {
		struct ar3cm_proto_analog *pkt =
			(struct ar3cm_proto_analog *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			return;
		ARecord *r = new ASingleRecord(ampi16_to_cpu(pkt->number),
				ampi16_to_cpu(pkt->value));
		if (CO->newOffARecord(amp64_to_cpu(pkt->timestamp), r))
			delete r;
		break;
	}
	case AR3CM_PROTO_RECORDDETECT: {
		struct ar3cm_proto_recorddetect *pkt =
			(struct ar3cm_proto_recorddetect *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;
		qWarning("detected %d analogs, %d binary",
			(int) amp16_to_cpu(pkt->a_len),
			(int) amp16_to_cpu(pkt->b_len) * 8);
		CO->hashSetupData(SetupSeq, SetupSeqOff);
		break;
	}
	case AR3CM_PROTO_BINARIES: {
		int i, n;
		struct ar3cm_proto_binaries *pkt =
			(struct ar3cm_proto_binaries *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;
		n = (size - sizeof(*pkt)) * 8;
		for (i = 0; i < n; i++)
			CO->newBRecord(lastTimestamp, i
					+ amp16_to_cpu(pkt->base),
					(amp8_to_cpu(pkt->data[i >> 3])
					& (1 << (i & 7))) ? High : Low);
		break;
	}
	case AR3CM_PROTO_BINARYLOW: {
		struct ar3cm_proto_binarylowhigh *rec =
			(struct ar3cm_proto_binarylowhigh *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*rec)))
			break;
		
		unsigned long long ts =	size == sizeof(*rec)
			? amp64_to_cpu(rec->timestamp) : lastTimestamp;
		CO->newBRecord(ts, amp16_to_cpu(rec->index), Low);
		break;
	}
	case AR3CM_PROTO_BINARYHIGH: {
		struct ar3cm_proto_binarylowhigh *rec =
			(struct ar3cm_proto_binarylowhigh *) buf;

		if (check_packet_size_min(hdr, sizeof(*rec)))
			break;

		unsigned long long ts =	size == sizeof(*rec)
			? amp64_to_cpu(rec->timestamp) : lastTimestamp;
		CO->newBRecord(ts, amp16_to_cpu(rec->index), High);
		break;
	}
	case AR3CM_PROTO_STOREDBLOW: {
		struct ar3cm_proto_binarylowhigh *rec =
			(struct ar3cm_proto_binarylowhigh *) buf;

		if (check_packet_size_min(hdr, sizeof(*rec)))
			break;
		
		unsigned long long ts =	size == sizeof(*rec)
			? amp64_to_cpu(rec->timestamp) : lastTimestamp;
		CO->newOffBRecord(ts, amp16_to_cpu(rec->index), Low);
		break;
	}
	case AR3CM_PROTO_STOREDBHIGH: {
		struct ar3cm_proto_binarylowhigh *rec =
			(struct ar3cm_proto_binarylowhigh *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*rec)))
			break;
		
		unsigned long long ts =	size == sizeof(*rec)
			? amp64_to_cpu(rec->timestamp) : lastTimestamp;
		CO->newOffBRecord(ts, amp16_to_cpu(rec->index), High);
		break;
	}
	case AR3CM_PROTO_VISSETUPSTART:
		qWarning("old protocol there, cannot continue...");
		//abort();
		break;
	case AR3CM_PROTO_NEWSETUPSTART:
		Owner->interface_reinit();
		break;
	case AR3CM_PROTO_ANALGRAPHSTART:
		Owner->newGraph();
		break;
	case AR3CM_PROTO_ANALHGRID: {
		struct ar3cm_proto_analhgrid *pkt =
			(struct ar3cm_proto_analhgrid *) buf;
		if (check_packet_size_min(hdr, sizeof(*pkt)
					+ amp8_to_cpu(pkt->nlines) * 2))
			break;

		/* TODO: zamienic na alloca lub uzyc lines[pkt->nlines] */
		int *lines = new int[amp8_to_cpu(pkt->nlines)];
		
		for (int i = 0; i < amp8_to_cpu(pkt->nlines); i++)
			lines[i] = (int) amp16_to_cpu(pkt->lines[i]);
		Owner->addHGrid(amp8_to_cpu(pkt->nlines), lines,
				hsv_qcolor(&pkt->color));
		delete[]lines;
		break;
	}
	case AR3CM_PROTO_ANALPLOT: {
		struct ar3cm_proto_analplot *pkt =
			(struct ar3cm_proto_analplot *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
			break;
		
		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->addPlot(amp16_to_cpu(pkt->physical),
				text2unicode(QString(pkt->caption)),
				hsv_qcolor(&pkt->color),
				amp_float_to_cpu(pkt->min),
				amp_float_to_cpu(pkt->max),
				amp_float_to_cpu(pkt->unit_value),
				text2unicode(QString(pkt->unit)));
		break;
	}
	case AR3CM_PROTO_ANALGRAPHEND:
		break;
	case AR3CM_PROTO_NEWTAB: {
		if (check_packet_size_min(hdr, 1))
			break;

		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->newTab(text2unicode(QString((const char *) buf)), false);
		break;
	}
	case AR3CM_PROTO_HIDDENTAB: {
		if (check_packet_size_min(hdr, 1))
			break;

		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->newTab(text2unicode(QString((const char *) buf)), true);
		break;
	}
	case AR3CM_PROTO_BINITEM:
		parse_binitem(hdr, buf);
		break;
	case AR3CM_PROTO_VISSETUPEND:
		FullyConnected = true;
		emit sigFullyConnected();
		Owner->visSetupEnd();
		break;
	case AR3CM_PROTO_NOP:
		break;
	case AR3CM_PROTO_VMETER: {
		struct ar3cm_proto_vmeter *pkt =
			(struct ar3cm_proto_vmeter *) buf;
		unsigned int size = amp16_to_cpu(hdr->size);
		unsigned int flags;
		char *p, *end;
		
		if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
			break;
		
		/* trim caption to not overflow */
		char *caption = strndup(pkt->caption, size - sizeof(*pkt));
		if (!caption)
			break;
		p = (char *)(pkt + 1);
		p += strlen(caption) + 1;
		end = (char *) pkt + size;

		flags = 0;
		if (p + sizeof(amp16_t) <= end)
			flags = amp16_to_cpu(get_unaligned((amp16_t *) p));
		buf[size - 1] = '\0';
		Owner->newVMeter(amp16_to_cpu(pkt->physical),
				text2unicode(QString(caption)),
				hsv_qcolor(&pkt->color),
				amp_float_to_cpu(pkt->min),
				amp_float_to_cpu(pkt->max),
				amp_float_to_cpu(pkt->unit_value),
				flags);
		free(caption);
		break;
	}
	case AR3CM_PROTO_TITLE: {
		struct ar3cm_proto_title *tit =
			(struct ar3cm_proto_title *) buf;
		
		if (check_packet_size_min(hdr, 1))
			break;
		
		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->setTitle(text2unicode(QString(tit->title)));
		break;
	}
	case AR3CM_PROTO_BINGRAPH: {
		struct ar3cm_proto_bingraph *pkt =
			(struct ar3cm_proto_bingraph *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;
		
		if (amp8_to_cpu(pkt->type) == AR3CM_BINGRAPH_DZWONEK)
			Owner->newDzwonek(amp16_to_cpu(pkt->physical),
				text2unicode(QString(pkt->symbol)),
				hsv_qcolor(&pkt->activecolor),
				hsv_qcolor(&pkt->inactivecolor));
		else
			Owner->newBinGraph(amp16_to_cpu(pkt->physical),
				text2unicode(QString(pkt->symbol)),
				hsv_qcolor(&pkt->activecolor),
				hsv_qcolor(&pkt->inactivecolor));
		CO->setBinaryUsed(amp16_to_cpu(pkt->physical));
		break;
	}
	case AR3CM_PROTO_LEFTBOX:
		Owner->newLeftbox();
		break;
	case AR3CM_PROTO_SCREENSAVERBOX:
		Owner->newScreenSaver();
		break;
	case AR3CM_PROTO_BOTTOMBOX:
		Owner->newBottombox();
		break;
	case AR3CM_PROTO_LABEL: {
		struct ar3cm_proto_label *pkt =
			(struct ar3cm_proto_label *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
			break;
		
		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->newLabel(text2unicode(QString(pkt->label)));
		break;
	}
	case AR3CM_PROTO_BIGLABEL: {
		struct ar3cm_proto_label *pkt =
			(struct ar3cm_proto_label *) buf;

		if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
			break;

		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->newBigLabel(text2unicode(QString(pkt->label)));
		break;
	}
	case AR3CM_PROTO_HBOX:
		Owner->newHBox();
		break;
	case AR3CM_PROTO_VBOX:
		Owner->newVBox();
		break;
	case AR3CM_PROTO_HLAYOUT:
		Owner->newHLayout();
		break;
	case AR3CM_PROTO_VLAYOUT:
		Owner->newVLayout();
		break;
	case AR3CM_PROTO_ENDLAYOUT:
		Owner->endLayout();
		break;
	case AR3CM_PROTO_SPACER:
		Owner->newSpacer();
		break;
	case AR3CM_PROTO_SIZEHINT:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->setSizeHint(__amp16_to_cpu(get_unaligned((__uint16_t *)
						buf)));
		break;
	case AR3CM_PROTO_TIMESCALE:
		Owner->newTimeScale();
		break;
	case AR3CM_PROTO_STRETCH:
		Owner->newStretch();
		break;
	case AR3CM_PROTO_SEPARATOR:
		Owner->newSeparator();
		break;
	case AR3CM_PROTO_BUTTON: {
		struct ar3cm_proto_button *pkt =
			(struct ar3cm_proto_button *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;
		Owner->newButton(text2unicode(QString(pkt->label)),
				parseAction((struct ar3cm_proto_action *)
					(pkt->label + strlen(pkt->label) + 1)));
		break;
	}
	case AR3CM_PROTO_SZYBOWSKAZ: {
		unsigned int i;
		struct ar3cm_proto_szybowskaz *pkt =
			(struct ar3cm_proto_szybowskaz *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)
					+ amp32_to_cpu(pkt->level_count)
					* sizeof(struct ar3cm_level)))
			break;
		for (i = 0; i < amp32_to_cpu(pkt->level_count); i++)
			pkt->level[i].caption[15] = '\0';
		Owner->newSzybowskaz(
				amp16_to_cpu(pkt->posphys),
				amp16_to_cpu(pkt->velphys),
				amp_float_to_cpu(pkt->posuv),
				amp_float_to_cpu(pkt->veluv),
				amp_float_to_cpu(pkt->minpos),
				amp_float_to_cpu(pkt->maxpos),
				amp_float_to_cpu(pkt->minvel),
				amp_float_to_cpu(pkt->maxvel),
				amp32_to_cpu(pkt->level_count), pkt->level,
				orientation & AR3CM_ORIENTATION_LEFT);
		break;
	}
	case AR3CM_PROTO_ENDOFDATA:
		CO->end_of_data();
		break;
	case AR3CM_PROTO_MASATEGOSYFU: {
		struct ar3cm_proto_masategosyfu *pkt =
			(struct ar3cm_proto_masategosyfu *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;
		CO->newSyf(amp64_to_cpu(pkt->calk_start),
				amp_float_to_cpu(pkt->akt_masa),
				amp_float_to_cpu(pkt->calk_masa));
		break;
	}
	case AR3CM_PROTO_UNDEF:	{
		struct ar3cm_proto_undef *pkt =
			(struct ar3cm_proto_undef *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;
		CO->beginUndefinedOffline(amp64_to_cpu(pkt->ts),
				amp32_to_cpu(pkt->flags));
		break;
	}
	case AR3CM_PROTO_AMETER: {
		struct ar3cm_proto_ameter *pkt =
			(struct ar3cm_proto_ameter *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
			break;
		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		pkt->unit[15] = '\0';
		Owner->newAMeter(amp16_to_cpu(pkt->physical),
				text2unicode(QString(pkt->caption)),
				hsv_qcolor(&pkt->color),
				amp_float_to_cpu(pkt->min),
				amp_float_to_cpu(pkt->max),
				amp_float_to_cpu(pkt->unit_value),
				ampi16_to_cpu(pkt->angle),
				ampi16_to_cpu(pkt->scale),
				text2unicode(QString(pkt->unit)),
				amp16_to_cpu(pkt->flags));
		break;
	}
	case AR3CM_PROTO_CLOCK: {
		struct ar3cm_proto_clock *pkt =
			(struct ar3cm_proto_clock *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
			break;
		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->newTClockLabel(pkt->format);
		break;
	}
	case AR3CM_PROTO_OBJECT: {
		struct ar3cm_proto_object *pkt =
			(struct ar3cm_proto_object *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt) + 1))
			break;
		/* trim caption to not overflow */
		buf[size - 1] = '\0';
		Owner->newObject(pkt->command);
		break;
	}
	case AR3CM_PROTO_DMETER: {
		struct ar3cm_proto_dmeter *pkt =
			(struct ar3cm_proto_dmeter *) buf;
		QString prefix = QString();

		if (check_packet_size_min(hdr, (char *)(&pkt->prefix)
					- (char *)pkt))
			break;
		if (!check_packet_size_min(hdr, sizeof(*pkt))) {
			const char *t = amp_string_get(&pkt->prefix, pkt, hdr);

			if (t != NULL)
				prefix = QString(t);
		}
		pkt->unit[15] = '\0';
		Owner->newDMeter(amp16_to_cpu(pkt->physical),
				hsv_qcolor(&pkt->color),
				amp_float_to_cpu(pkt->unit_value),
				text2unicode(QString(pkt->unit)),
				amp16_to_cpu(pkt->flags),
				prefix);
		break;
	}
	case AR3CM_PROTO_MINWIDTH:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->setMinWidthAttr(__amp16_to_cpu(get_unaligned(
						(__uint16_t *) buf)));
		break;
	case AR3CM_PROTO_MAXWIDTH:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->setMaxWidthAttr(__amp16_to_cpu(get_unaligned(
						(__uint16_t *) buf)));
		break;
	case AR3CM_PROTO_MINHEIGHT:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->setMinHeightAttr(__amp16_to_cpu(get_unaligned(
						(__uint16_t *) buf)));
		break;
	case AR3CM_PROTO_MAXHEIGHT:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->setMaxHeightAttr(__amp16_to_cpu(get_unaligned(
						(__uint16_t *) buf)));
		break;
	case AR3CM_PROTO_STRETCHA:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->setStretchAttr(__amp16_to_cpu(get_unaligned(
						(__uint16_t *) buf)));
		break;
	case AR3CM_PROTO_VERSION_PKT: {
		if (check_packet_size_min(hdr, 4))
			break;
		unsigned int ver = __amp32_to_cpu(get_unaligned((__uint32_t *)
					buf));
		
		qWarning("protocol version: local: %x, remote: %x",
				AR3CM_PROTO_VERSION, ver);
		break;
	}
	case AR3CM_PROTO_SCHEME_ID: {
		if (check_packet_size_min(hdr, 4))
			break;
		
		qWarning("scheme: %d", (unsigned int) __amp32_to_cpu(
					get_unaligned((__uint32_t *) buf)));
		break;
	}
	case AR3CM_PROTO_ENCODING:
		config_encoding = strndup((const char *) buf, size);
		qWarning("encoding: %s\n", config_encoding);
		break;
	
	case AR3CM_PROTO_ORIENTATION:
		if (check_packet_size_eq(hdr, 2))
			break;
		orientation = __amp16_to_cpu(get_unaligned((__uint16_t *) buf));
		break;
	
	case AR3CM_PROTO_TOKEN:
		receive_token(hdr, (struct ar3cm_proto_token *) buf);
		break;
	
	case AR3CM_PROTO_PENDRIVE_NOTIFY: {
		struct ar3cm_proto_pendrive_notify *pkt =
			(struct ar3cm_proto_pendrive_notify *) buf;
		unsigned int events;
		unsigned int status;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;

		events = amp32_to_cpu(pkt->events);
		status = amp32_to_cpu(pkt->status);
		
		if (events & AR3CM_PROTO_PENDRIVE_POLL_INSERT)
			emit sig_pendrive_insert();
		if (events & AR3CM_PROTO_PENDRIVE_POLL_EJECT)
			emit sig_pendrive_eject();
		if (events & AR3CM_PROTO_PENDRIVE_POLL_COMPLETE)
			emit sig_pendrive_complete(status);
		else if (events & AR3CM_PROTO_PENDRIVE_POLL_ERR)
			emit sig_pendrive_error(status);
		if (events & AR3CM_PROTO_PENDRIVE_POLL_PROGRESS)
			emit sig_pendrive_progress(status);
		break;
	}

	case AR3CM_PROTO_EMPTY: {
		Owner->newEmpty();
		break;
	}

	case AR3CM_PROTO_SETCOLOR:
	case AR3CM_PROTO_SETBGCOLOR: {
		struct ar3cm_proto_setcolor *pkt =
			(struct ar3cm_proto_setcolor *) buf;
		
		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;

		QColor color(hsv_qcolor(&pkt->color));
		if (type == AR3CM_PROTO_SETCOLOR)
			Owner->setColor(color);
		else
			Owner->setBgColor(color);
		break;
	}
	case AR3CM_PROTO_SET_FONT_SIZE:
	case AR3CM_PROTO_SET_VALUE_FONT_SIZE:
	case AR3CM_PROTO_SET_TIMESCALE_FONT_SIZE:
	case AR3CM_PROTO_SET_SMALL_FONT_SIZE: {
		if (check_packet_size_min(hdr, 4))
			break;
		unsigned int font = __amp32_to_cpu(get_unaligned((__uint32_t *)
					buf));
		unsigned int base_size;
		unsigned int size;
		unsigned int ftype = font & AR3CM_PROTO_FONT_SIZE_MASK;
		unsigned int p = font & ~AR3CM_PROTO_FONT_SIZE_MASK;
		unsigned int i;
		
		base_size = ar3cmon_font_size(AR3CMONITOR_FONT_BASE);
		if (ftype == AR3CM_PROTO_FONT_SIZE_PX) {
			size = p;
		} else if (ftype == AR3CM_PROTO_FONT_SIZE_MAGNIFY) {
			size = base_size * p / 1000;
		} else if (ftype == AR3CM_PROTO_FONT_SIZE_PLUS) {
			size = base_size * 1000;
			if (p >= 11 && p <= 19) {
				for (i = 10; i < p; i++)
					size = size * 12 / 10;
			} else if ((p >= 1) && (p < 10)) {
				for (i = 10; i > p; i--)
					size = size * 10 / 12;
			}
			size /= 1000;
		} else {
			size = base_size;
		}

		if (type == AR3CM_PROTO_SET_VALUE_FONT_SIZE) {
			CO->setValueFontSize(size);
		} else if (type == AR3CM_PROTO_SET_FONT_SIZE) {
			CO->setFontSize(size);
		} else if (type == AR3CM_PROTO_SET_TIMESCALE_FONT_SIZE) {
			CO->setTimescaleFontSize(size);
		} else if (type == AR3CM_PROTO_SET_SMALL_FONT_SIZE) {
			CO->setSmallFontSize(size);
		}
		break;
	}
	case AR3CM_PROTO_SET_PREC: {
		if (check_packet_size_min(hdr, 4))
			break;
		unsigned int prec = __amp32_to_cpu(get_unaligned((__uint32_t *)
					buf));
		unsigned int t = prec & AR3CM_PROTO_PREC_MASK
				& ~AR3CM_PROTO_PREC_LZERO;
		unsigned int d = prec & ~AR3CM_PROTO_PREC_MASK;
		prec_t p = PREC_SDIGITS;

		if (t == AR3CM_PROTO_PREC_DDIGITS)
			p &= ~PREC_SDIGITS;

		if (prec & AR3CM_PROTO_PREC_LZERO)
			p |= PREC_LZERO;

		p |= d & PREC_DIGITS_MASK;
		
		CO->set_prec(p);
		break;
	}
	case AR3CM_PROTO_TCN: {
		struct ar3cm_proto_tcn *pkt =
			(struct ar3cm_proto_tcn *) buf;

		if (check_packet_size_min(hdr, sizeof(*pkt)))
			break;

		if (amp64_to_cpu(pkt->new_ts) < CO->timeStamp()) {
			lastTimestamp = amp64_to_cpu(pkt->new_ts) - 1;
			CO->clearOnlineData();
		} else {
			if (amp64_to_cpu(pkt->old_ts) + 1 > CO->timeStamp()) {
				CO->beginUndefinedRegion(
						amp64_to_cpu(pkt->old_ts) + 1);
			} else if (!amp64_to_cpu(pkt->old_ts)) {
				CO->beginUndefinedRegion(CO->timeStamp() + 1);
			} else {
				/*
				 * something is wrong, it's a good idea to
				 * clear all online data
				 */
				CO->clearOnlineData();
				lastTimestamp = amp64_to_cpu(pkt->new_ts) - 1;
			}
		}
		break;
	}
	case AR3CM_PROTO_HIDE_IF:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->hide_cond(__amp16_to_cpu(get_unaligned(
						(__uint16_t *) buf)), 0);
		break;
	case AR3CM_PROTO_HIDE_UNLESS:
		if (check_packet_size_eq(hdr, 2))
			break;
		Owner->hide_cond(__amp16_to_cpu(get_unaligned(
						(__uint16_t *) buf)), 1);
		break;
	case AR3CM_PROTO_STATUS: {
		if (check_packet_size_min(hdr, 4))
			break;

		__uint32_t *st = (__uint32_t *) buf;
		unsigned int status = __amp32_to_cpu(get_unaligned(st));

		if (status)
			qWarning("remote command error %d: %s", status,
					strerror(status));
		break;
	}
	default: {
		qWarning("unrecognized packet ignored: %02x", type);
		break;
	}
	}
}

Action *CommObject::parseAction(struct ar3cm_proto_action *a)
{
	/* BUG: attaching actions to Owner is not a good idea,
	 * because these actions shoud be freed on every disconnect. */
	switch (amp16_to_cpu(a->type)) {
	case AR3CM_ACTION_QUIT:
		return new AcQuit(Owner);
	case AR3CM_ACTION_DEBUG:
		return new AcDebug((char *) (a + 1), Owner);
	case AR3CM_ACTION_CHTAB:
		return new AcChTab(amp16_to_cpu(((struct ar3cm_proto_chtab *)
						(a + 1))->tab), Owner);
	case AR3CM_ACTION_OFFLINE: {
		AcSignal *a = new AcSignal(Owner);
		connect(a, SIGNAL(sig()), CO, SLOT(enterOfflineMode()));
		return a;
	}
	case AR3CM_ACTION_REVIEW: {
		AcSignal *a = new AcSignal(Owner);
		connect(a, SIGNAL(sig()), CO, SLOT(enterReviewMode()));
		return a;
	}
	case AR3CM_ACTION_NORMAL: {
		AcSignal *a = new AcSignal(Owner);
		connect(a, SIGNAL(sig()), CO, SLOT(enterNormalMode()));
		return a;
	}
	case AR3CM_ACTION_DISCONNECT: {
		AcSignal *a = new AcSignal(Owner);
		connect(a, SIGNAL(sig()), this, SLOT(goOffline()));
		return a;
	}
	case AR3CM_ACTION_CONNECT: {
		AcSignal *a = new AcSignal(Owner);
		connect(a, SIGNAL(sig()), this, SLOT(connectToAr3c()));
		return a;
	}
	default:
		qWarning("unrecognized action, substituting with AcDebug");
		return new AcDebug("unrecognized action called!", Owner);
	}
}

void CommObject::handleSetupSeq(unsigned char *buf)
{
	if (SetupSeqOff == -1)
		return;

	switch (amp16_to_cpu(PacketHdr.type)) {
		case AR3CM_PROTO_AMETER:
		case AR3CM_PROTO_ANALGRAPHEND:
		case AR3CM_PROTO_ANALGRAPHSTART:
		case AR3CM_PROTO_ANALHGRID:
		case AR3CM_PROTO_ANALPLOT:
		case AR3CM_PROTO_BIGLABEL:
		case AR3CM_PROTO_BINGRAPH:
		case AR3CM_PROTO_BINITEM:
		case AR3CM_PROTO_BOTTOMBOX:
		case AR3CM_PROTO_BUTTON:
		case AR3CM_PROTO_DMETER:
		case AR3CM_PROTO_EMPTY:
		case AR3CM_PROTO_ENDLAYOUT:
		case AR3CM_PROTO_HBOX:
		case AR3CM_PROTO_HIDDENTAB:
		case AR3CM_PROTO_HLAYOUT:
		case AR3CM_PROTO_LABEL:
		case AR3CM_PROTO_LEFTBOX:
		case AR3CM_PROTO_MAXHEIGHT:
		case AR3CM_PROTO_MAXWIDTH:
		case AR3CM_PROTO_MINHEIGHT:
		case AR3CM_PROTO_MINWIDTH:
		case AR3CM_PROTO_NEWSETUPSTART:
		case AR3CM_PROTO_NEWTAB:
		case AR3CM_PROTO_ORIENTATION:
		case AR3CM_PROTO_SCREENSAVERBOX:
		case AR3CM_PROTO_SEPARATOR:
		case AR3CM_PROTO_SETBGCOLOR:
		case AR3CM_PROTO_SETCOLOR:
		case AR3CM_PROTO_SIZEHINT:
		case AR3CM_PROTO_SPACER:
		case AR3CM_PROTO_STRETCH:
		case AR3CM_PROTO_STRETCHA:
		case AR3CM_PROTO_SZYBOWSKAZ:
		case AR3CM_PROTO_TIMESCALE:
		case AR3CM_PROTO_TITLE:
		case AR3CM_PROTO_VBOX:
		case AR3CM_PROTO_VISSETUPEND:
		case AR3CM_PROTO_VLAYOUT:
		case AR3CM_PROTO_VMETER:
			break;
		default:
			return;
	}

	if (SetupSeqOff + sizeof(PacketHdr) + amp16_to_cpu(PacketHdr.size)
			>= AR3CMON_SETUPSEQ_BUF) {
		qWarning("setup sequence buffer too small, "
			"offline mode not available)");
		SetupSeqOff = -1;
	}

	memcpy(SetupSeq + SetupSeqOff, &PacketHdr, sizeof(PacketHdr));
	SetupSeqOff += sizeof(PacketHdr);
	memcpy(SetupSeq + SetupSeqOff, buf, amp16_to_cpu(PacketHdr.size));
	SetupSeqOff += amp16_to_cpu(PacketHdr.size);
}

void CommObject::startOfflineProtocol()
{
	long i = 0;
	qWarning("replaying startup sequence in offline mode ...");
	emit sigConnecting();
	emit sigOfflineConnected();
	while (i < SetupSeqOff) {
		struct ar3cm_proto_hdr *h =
			(struct ar3cm_proto_hdr *) (SetupSeq + i);
		unsigned char *buf = (unsigned char *) (h + 1);
		parsePacket(h, buf);
		i += sizeof(*h) + amp16_to_cpu(h->size);
	}
	qWarning("... done");
	CO->enterReviewMode();
}

void CommObject::watchdogHandler()
{
	qWarning("Watchdog trigger");
	errorHandler(-5);
}

void CommObject::requestScheme(unsigned int scheme)
{
	amp32_t pkt;
	
	pkt = cpu_to_amp32(scheme);
	sendCommand(AR3CM_PROTO_SETUP_INIT, &pkt, sizeof(pkt));
}

void CommObject::request_locale(const char *locale)
{
	sendCommand(AR3CM_PROTO_SETLOCALE, locale, strlen(locale) + 1);
}

void CommObject::set_opt(unsigned long opt, unsigned long value)
{
	struct ar3cm_proto_set_opt set_opt;

	set_opt.opt = cpu_to_amp32(opt);
	set_opt.value = cpu_to_amp32(value);
	sendCommand(AR3CM_PROTO_SET_OPT, &set_opt, sizeof(set_opt));
}

void CommObject::requestMinute(unsigned long min)
{
	struct ar3cm_proto_getminute pkt;

	pkt.min = cpu_to_amp32(min);
	pkt.layer = cpu_to_ampi32(CO->get_layer());
	sendCommand(AR3CM_PROTO_GETMINUTE, &pkt, sizeof(pkt));
}

void CommObject::requestData()
{
	sendSimpleCommand(AR3CM_PROTO_SENDBINARIES);
}
	
int CommObject::setDataloggerTime(unsigned long long ts, bool force_imidiate)
{
	struct ar3cm_proto_set_time *pkt;
	struct signature s;
	int ret;
	int s_size;

	if (token_state != TOKEN_READY)
		return -EAGAIN;

	ret = signature_init(&s);
	if (ret)
		goto out;

	ret = signature_set_token(&s, &token);
	if (ret)
		goto out_destroy;

	ret = signature_reset(&s);
	if (ret)
		goto out_destroy;

	ret = signature_size(&s);
	if (ret < 0)
		goto out_destroy;

	s_size = ret;
	pkt = (typeof(pkt)) malloc(sizeof(*pkt) + s_size);
	ret = -ENOMEM;
	if (pkt == NULL)
		goto out_destroy;

	pkt->ts = cpu_to_amp64(ts);
	pkt->flags = force_imidiate ? cpu_to_amp32(0)
		: cpu_to_amp32(AR3CM_PROTO_MAY_ADJUST);

	ret = signature_update(&s, pkt, sizeof(*pkt) - sizeof(pkt->signature));
	if (ret)
		goto out_free;

	ret = signature_update(&s, password, password_size);
	if (ret)
		goto out_free;

	ret = signature_generate(&s, pkt + 1);
	if (ret)
		goto out_free;

	ar3cm_dblock_set(&pkt->signature.sign, sizeof(*pkt), s_size);

	sendCommand(AR3CM_PROTO_SET_TIME, pkt, sizeof(*pkt) + s_size);
out_free:
	free(pkt);
out_destroy:
	signature_destroy(&s);
out:
	return ret;
}

int CommObject::pendrive_copy(unsigned long long from,
		unsigned long long to, int format)
{
	struct ar3cm_proto_pendrive_copy *pkt;
	struct signature s;
	int ret;
	int s_size;

	if (token_state != TOKEN_READY)
		return -EAGAIN;

	ret = signature_init(&s);
	if (ret)
		goto out;

	ret = signature_set_token(&s, &token);
	if (ret)
		goto out_destroy;

	ret = signature_reset(&s);
	if (ret)
		goto out_destroy;

	ret = signature_size(&s);
	if (ret < 0)
		goto out_destroy;

	s_size = ret;
	pkt = (typeof(pkt)) malloc(sizeof(*pkt) + s_size);
	ret = -ENOMEM;
	if (pkt == NULL)
		goto out_destroy;

	pkt->from = cpu_to_amp64(from);
	pkt->to = cpu_to_amp64(to);
	pkt->format = cpu_to_amp16(format);

	ret = signature_update(&s, pkt, sizeof(*pkt) - sizeof(pkt->signature));
	if (ret)
		goto out_free;

	ret = signature_update(&s, password, password_size);
	if (ret)
		goto out_free;

	ret = signature_generate(&s, pkt + 1);
	if (ret)
		goto out_free;

	ar3cm_dblock_set(&pkt->signature.sign, sizeof(*pkt), s_size);

	sendCommand(AR3CM_PROTO_PENDRIVE_COPY, pkt, sizeof(*pkt) + s_size);
out_free:
	free(pkt);
out_destroy:
	signature_destroy(&s);
out:
	return ret;
			
	return 0;
}

void CommObject::get_token(void)
{
	struct ar3cm_proto_gettoken pkt;

	pkt.types = cpu_to_amp32(AR3CM_TOKEN_SHA1);
	pkt.flags = cpu_to_amp32(0);
	pkt.ttl = cpu_to_amp64(TOKEN_TTL);
	sendCommand(AR3CM_PROTO_GETTOKEN, &pkt, sizeof(pkt));
	token_state = TOKEN_WANT;
}

int CommObject::set_password(const char *pass)
{
#ifdef HAVE_SSL
	long page_size = sysconf(_SC_PAGESIZE);
	void *_context;
	SHA_CTX *context;
	void *digest;
	int ret;
       	
	ret = posix_memalign(&_context, page_size, sizeof(*context));
	if (ret)
		goto err;
	context = (SHA_CTX *) _context;

	ret = posix_memalign(&digest, page_size, SHA_DIGEST_LENGTH);
	if (ret)
		goto err_context;

	mlock(digest, page_size);
	mlock(context, page_size);

	SHA1_Init(context);
	SHA1_Update(context, pass, strlen(pass));
	SHA1_Final((unsigned char *) digest, context);

	password = (typeof(password)) digest;
	password_size = SHA_DIGEST_LENGTH;

err_context:
	free(context);
err:
	return ret;
#else
	(void) pass;
	return 0;
#endif
}
