/*
 * ar3cmonitor - datastore.cpp
 *
 * 	Data Store
 * 
 * Copyright (C) 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ar3cmonitor.h"
#include "datastore.h"
#include "control.h"
#include <stdio.h>
#include <stdarg.h>

extern unsigned long stores;
		
DataStore::DataStore(unsigned int binaries, const char *_name)
	: nBinaries(binaries), used(0), records(0), CursorStamp(0ULL),
	beg_ts(0ULL), end_ts(0ULL)
{
	Analog = new AnalogControl;
	Binary = new BinaryControl[nBinaries];
	stores++;
	name = (_name != NULL) ? _name : "unnamed";
}
	
void DataStore::error(const char *fmt, ...)
{
	va_list ap;
	
	fprintf(stderr, "%s: ", name);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

void DataStore::update_ts(unsigned long long ts)
{
	if (records) {
		if (ts < beg_ts)
			beg_ts = ts;
		else if (ts > end_ts)
			end_ts = ts;
	} else {
		beg_ts = ts;
		end_ts = ts;
	}
}

void DataStore::insert_BRecord(unsigned long long ts, int id, BRecord *r)
{
	BinaryControl *c = &Binary[id];
	c->InsertIt = c->InsertIt.insertAfter(ts, r);
	if (c->Cursor.timeStamp() <= CursorStamp)
		++(c->Cursor);
	update_ts(ts);
	stat_new_record(monalloc_tsize(r->get_size()));
}

void DataStore::newBRecord(unsigned long long ts, int id, BState state)
{
	BinaryControl *c = get_binary(id);
	if (c == NULL)
		return;

	BRecord *r, *prev = c->InsertIt.data();
	BState ps = (prev && prev->valid) ? prev->state : Undefined;
	
	if (ps == state)
		return;
	
	r = new BRecord();
	r->state = state;
	insert_BRecord(ts, id, r);
	
	garbageCollector();
}

void DataStore::insert_ARecord(unsigned long long ts, ARecord *r)
{
	Analog->InsertIt = Analog->InsertIt.insertAfter(ts, r);
	if (Analog->Cursor.timeStamp() <= CursorStamp)
		++(Analog->Cursor);
	update_ts(ts);
	stat_new_record(monalloc_tsize(r->get_size()));
}

void DataStore::newARecord(unsigned long long ts, ARecord *r)
{
	insert_ARecord(ts, r);
	garbageCollector();
}

void DataStore::seek_eof()
{
	unsigned int i;
	
	Analog->InsertIt = Analog->Data.it() - 1;
	for (i = 0; i < nBinaries; i++)
		Binary[i].InsertIt = Binary[i].Data.it() - 1;
}

void DataStore::cursors_eof()
{
	unsigned int i;
	
	Analog->Cursor = Analog->Data.it() - 1;
	for (i = 0; i < nBinaries; i++)
		Binary[i].Cursor = Binary[i].Data.it() - 1;
}

void DataStore::rewind()
{
	unsigned int i;
	
	Analog->InsertIt = Analog->Data.it();
	for (i = 0; i < nBinaries; i++)
		Binary[i].InsertIt = Binary[i].Data.it();
}

int DataStore::shrink(int)
{
	return 0;
}

int DataStore::left_join(DataStore &a)
{
	unsigned int i;
	
	Analog->Data.left_join(a.Analog->Data);
	for (i = 0; i < nBinaries; i++)
		Binary[i].Data.left_join(a.Binary[i].Data);
	beg_ts = a.beg_ts;
	records += a.get_records();
	used += a.get_used();
	return 0;
}

int DataStore::right_join(DataStore &a)
{
	unsigned int i;

	Analog->Data.right_join(a.Analog->Data);
	for (i = 0; i < nBinaries; i++)
		Binary[i].Data.right_join(a.Binary[i].Data);
	end_ts = a.end_ts;
	records += a.get_records();
	used += a.get_used();
	return 0;
}

int DataStore::join(DataStore &a)
{
	if (a.empty())
		return 0;

#ifdef DATASTORE_STATS
	qWarning("join: b = %Lu, e = %Lu, a.beg_ts = %Lu a.end_ts = %Lu "
			"%lu vs. %lu",
			TS_MIN(beg_ts), TS_MIN(end_ts),
			TS_MIN(a.beg_ts), TS_MIN(a.end_ts),
			get_records(), a.get_records());
#endif
	if (empty()) {
		left_join(a);
		end_ts = a.end_ts;
		return 0;
	}

	if (a.beg_ts > end_ts)
		return right_join(a);
	if (beg_ts > a.end_ts)
		return left_join(a);
	return 1;
}

bool DataStore::empty()
{
	return records == 0;
}

void DataStore::clear()
{
	unsigned int i;

	Analog->clear();
	for (i = 0; i < nBinaries; i++)
		Binary[i].clear();
	records = 0;
	used = 0;
}

DataStore::~DataStore()
{
	clear();
	stores--;
	delete Analog;
	delete []Binary;
}
