/*
 * ar3cmonitor - pendrive.cpp
 *
 * 	Pendrive dialog
 *
 * Copyright (C) 2008  Krzysztof Kasperski <casper@podlesie.net>
 * Copyright (C) 2006..2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ar3cmonitor.h"
#include "consts.h"
#include "commobj.h"
#include "ar3cm_proto.h"

#include <qlabel.h>
#include <qpainter.h>
#include <qmessagebox.h>

#if QT_VERSION >= 0x40000
#define setProgress(x)		setValue(x)
#define totalSteps()		maximum()
#define setTotalSteps(x)	setMaximum(x)
#endif

#include <qstring.h>

#include "pendrive.h"

PenDrive::PenDrive(unsigned long long time, QWidget *parent)
	: QDialog(parent), pb(NULL)
{
	int pen_connected = CO->penDriveConnected();
	QDateTime t;

	QFont font(CO->hugeFont());

	t.setTime_t(time / 1000000000ULL);
	t = t.addMonths(-1);

	vl = new QVBoxLayout(this, 8);

	lb0 = new QLabel(trUtf8("<center>Backup dataloger's storage to pendrive."
			"</center>"), this);
	vl->addWidget(lb0);
	lb1 = new QLabel(trUtf8("<center>Choose time range to save.</center>"), this);
	vl->addWidget(lb1);

	h0 = new QHBoxLayout(vl);

	from_lb = new QLabel(trUtf8("<b>From:</b>"), this);
	h0->addWidget(from_lb);

	from = new QDateTimeEdit(t, this);
	from->setFont(font);
#if QT_VERSION >= 0x40000
	from->setDisplayFormat("dd-MM-yyyy hh:mm:ss");
#else
	from->dateEdit()->setOrder(QDateEdit::YMD);
#endif
	/* Qt 3.3.3b incorect size bug workaround */
	from->setMinimumWidth(from->sizeHint().width()
			+ from->fontMetrics().width("999"));
	h0->addWidget(from);

	if (!pen_connected)
		from->setEnabled(false);
	from->show();

	h1 = new QHBoxLayout(vl);

	till_lb = new QLabel(trUtf8("<b>Till:</b>"), this);
	h1->addWidget(till_lb);

	t = t.addMonths(1);
	till = new QDateTimeEdit(t, this);
	till->setFont(font);
#if QT_VERSION >= 0x40000
	till->setDisplayFormat("dd-MM-yyyy hh:mm:ss");
#else
	till->dateEdit()->setOrder(QDateEdit::YMD);
#endif
	/* Qt 3.3.3b incorect size bug workaround */
	till->setMinimumWidth(till->sizeHint().width()
			+ till->fontMetrics().width("999"));
	h1->addWidget(till);
	if (!pen_connected)
		till->setEnabled(false);
	till->show();

	if (!pen_connected) {
		warning = new QLabel(trUtf8("<center><font style='color: red;'>"
				"Pendrive is not connected. "
				"Connect your USB pendrive to the USB socket "
				"and try again.</font></center>"), this);
		vl->addWidget(warning);
	}

	h2 = new QHBoxLayout(vl);

	format_lb = new QLabel(trUtf8("Select data format: "), this);
	h2->addWidget(format_lb);

	format_cb = new QComboBox(this);
	format_cb->insertItem("Logger4");
	format_cb->insertItem("Logger3");
	format_cb->insertItem("Logger2");
	format_cb->setFont(font);
	h2->addWidget(format_cb);

	hl = new QHBoxLayout(vl);

	start = new QPushButton(tr("Start"), this);
	if (!pen_connected)
		start->setEnabled(false);
	connect(start, SIGNAL(clicked()), this, SLOT(start_clicked()));
	hl->addWidget(start);
	start->show();

	cancel = new QPushButton(trUtf8("Cancel"), this);
	connect(cancel, SIGNAL(clicked()), this, SLOT(cancel_clicked()));
	hl->addWidget(cancel);
	cancel->show();

}

void PenDrive::start_clicked()
{
	if (from->dateTime().toTime_t() >= till->dateTime().toTime_t()) {
		QMessageBox *mb = new QMessageBox(trUtf8("Warning"),
				trUtf8("<p>Date specified in 'till' window is "
				"earlier than the one in 'from' window</p>"),
				QMessageBox::Warning, QMessageBox::Ok,
				QMessageBox::NoButton, QMessageBox::NoButton,
				this);
		mb->show();
		return;
	}

	int mins_todo = (till->dateTime().toTime_t()
			- from->dateTime().toTime_t()) / 60;
	if (!pb) {
		pb = new QProgressBar(this);
		vl->addWidget(pb);
	}
	pb->setTotalSteps(mins_todo);
	pb->show();
	pb->setProgress(0);
	start->setEnabled(false);

	connect(TheCommObject, SIGNAL(sig_pendrive_insert()),
			SLOT(pendrive_inserted()));
	connect(TheCommObject, SIGNAL(sig_pendrive_eject()),
			SLOT(pendrive_ejected()));
	connect(TheCommObject, SIGNAL(sig_pendrive_progress(unsigned int)),
			SLOT(pendrive_progressed(unsigned int)));
	connect(TheCommObject, SIGNAL(sig_pendrive_complete(unsigned int)),
			SLOT(pendrive_completed(unsigned int)));

	int format = AR3CM_PROTO_PENDRIVE_FORMAT_LOGGER4;

	switch (format_cb->currentItem()) {
		case 0:
			format = AR3CM_PROTO_PENDRIVE_FORMAT_LOGGER4;
			break;
		case 1:
			format = AR3CM_PROTO_PENDRIVE_FORMAT_LOGGER3;
			break;
		case 2:
			format = AR3CM_PROTO_PENDRIVE_FORMAT_LOGGER2;
			break;
		default:
			format = AR3CM_PROTO_PENDRIVE_FORMAT_LOGGER4;
	}

	TheCommObject->pendrive_copy(1000000000ULL
			* from->dateTime().toTime_t(),
			1000000000ULL * till->dateTime().toTime_t(),
			format);
}

void PenDrive::pendrive_inserted()
{
	QMessageBox *mb = new QMessageBox(trUtf8("Warning"), trUtf8("Pendrive inserted"),
			QMessageBox::Warning, QMessageBox::Ok,
			QMessageBox::NoButton, QMessageBox::NoButton,
			this);
	mb->show();
	start->setEnabled(true);
}


void PenDrive::pendrive_ejected()
{
	QMessageBox *mb = new QMessageBox(trUtf8("Warning"), trUtf8("Pendrive ejected"),
			QMessageBox::Warning, QMessageBox::Ok,
			QMessageBox::NoButton, QMessageBox::NoButton,
			this);
	mb->show();
	start->setEnabled(false);
}

void PenDrive::pendrive_progressed(unsigned int progress)
{
	pb->setProgress(progress);
}

void PenDrive::pendrive_completed(unsigned int status)
{
	pb->setProgress(pb->totalSteps());
	start->setEnabled(true);
	QMessageBox *mb;

	if (!status) {
		mb = new QMessageBox(trUtf8("Ready"), trUtf8("Backup completed"),
				QMessageBox::Information, QMessageBox::Ok,
				QMessageBox::NoButton, QMessageBox::NoButton,
				this);
	} else {
		mb = new QMessageBox(trUtf8("Error"), trUtf8("Backup failed"),
				QMessageBox::Warning, QMessageBox::Ok,
				QMessageBox::NoButton, QMessageBox::NoButton,
				this);
	}
	mb->show();
}

void PenDrive::cancel_clicked()
{
	close();
}

PenDrive::~PenDrive()
{
	delete cancel;
	delete start;
	delete from;
	delete till;
	delete hl;
	delete vl;
}
